# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
from flask import Blueprint, request
from flask_apispec import marshal_with, use_kwargs

from renku.service.config import SERVICE_PREFIX
from renku.service.controllers.cache_files_upload import UploadFilesCtrl
from renku.service.controllers.cache_list_projects import ListProjectsCtrl
from renku.service.controllers.cache_list_uploaded import ListUploadedFilesCtrl
from renku.service.controllers.cache_migrate_project import MigrateProjectCtrl
from renku.service.controllers.cache_migrations_check import MigrationsCheckCtrl
from renku.service.controllers.cache_project_clone import ProjectCloneCtrl
from renku.service.serializers.cache import (
    FileListResponseRPC,
    FileUploadRequest,
    FileUploadResponseRPC,
    ProjectCloneRequest,
    ProjectCloneResponseRPC,
    ProjectListResponseRPC,
    ProjectMigrateRequest,
    ProjectMigrateResponseRPC,
    ProjectMigrationCheckRequest,
    ProjectMigrationCheckResponseRPC,
)
from renku.service.views.decorators import (
    accepts_json,
    handle_common_except,
    handle_migration_except,
    header_doc,
    requires_cache,
    requires_identity,
)

CACHE_BLUEPRINT_TAG = "cache"
cache_blueprint = Blueprint("cache", __name__, url_prefix=SERVICE_PREFIX)


@marshal_with(FileListResponseRPC)
@header_doc(description="List uploaded files.", tags=(CACHE_BLUEPRINT_TAG,))
@cache_blueprint.route(
    "/cache.files_list", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_uploaded_files_view(user_data, cache):
    """List uploaded files ready to be added to projects."""
    return ListUploadedFilesCtrl(cache, user_data).to_response()


@use_kwargs(FileUploadRequest)
@marshal_with(FileUploadResponseRPC)
@header_doc(
    description="Upload file or archive of files.", tags=(CACHE_BLUEPRINT_TAG,),
)
@cache_blueprint.route(
    "/cache.files_upload", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def upload_file_view(user_data, cache):
    """Upload file or archive of files."""
    return UploadFilesCtrl(cache, user_data, request).to_response()


@use_kwargs(ProjectCloneRequest)
@marshal_with(ProjectCloneResponseRPC)
@header_doc(
    "Clone a remote project. If the project is cached already, "
    "new clone operation will override the old cache state.",
    tags=(CACHE_BLUEPRINT_TAG,),
)
@cache_blueprint.route(
    "/cache.project_clone", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def project_clone_view(user_data, cache):
    """Clone a remote repository."""
    return ProjectCloneCtrl(cache, user_data, dict(request.json)).to_response()


@marshal_with(ProjectListResponseRPC)
@header_doc(
    "List cached projects.", tags=(CACHE_BLUEPRINT_TAG,),
)
@cache_blueprint.route(
    "/cache.project_list", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
def list_projects_view(user_data, cache):
    """List cached projects."""
    return ListProjectsCtrl(cache, user_data).to_response()


@use_kwargs(ProjectMigrateRequest)
@marshal_with(ProjectMigrateResponseRPC)
@header_doc(
    "Migrate project to the latest version.", tags=(CACHE_BLUEPRINT_TAG,),
)
@cache_blueprint.route(
    "/cache.migrate", methods=["POST"], provide_automatic_options=False,
)
@handle_common_except
@handle_migration_except
@accepts_json
@requires_cache
@requires_identity
def migrate_project_view(user_data, cache):
    """Migrate specified project."""
    return MigrateProjectCtrl(cache, user_data, dict(request.json)).to_response()


@use_kwargs(ProjectMigrationCheckRequest)
@marshal_with(ProjectMigrationCheckResponseRPC)
@header_doc(
    "Check if project requires migration.", tags=(CACHE_BLUEPRINT_TAG,),
)
@cache_blueprint.route(
    "/cache.migrations_check", methods=["GET"], provide_automatic_options=False,
)
@handle_common_except
@requires_cache
@requires_identity
@accepts_json
def migration_check_project_view(user_data, cache):
    """Migrate specified project."""
    return MigrationsCheckCtrl(cache, user_data, dict(request.args)).to_response()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Communicator class for printing click output."""

import sys

import click
from tqdm import tqdm

from renku.core.utils.communication import CommunicationCallback


class StandardOutput(CommunicationCallback):
    """Communication listener that outputs to stdout/stderr."""

    def __init__(self):
        super().__init__()
        self._progress_bars = {}
        self._progress_types = ["download"]

    def echo(self, msg):
        """Write a message."""
        with CommunicationCallback.lock:
            print(msg)

    def info(self, msg):
        """Write an info message."""
        with CommunicationCallback.lock:
            print(msg)

    def warn(self, msg):
        """Write a warning message."""
        with CommunicationCallback.lock:
            print(msg)

    def error(self, msg):
        """Write an error message."""
        with CommunicationCallback.lock:
            print(msg, file=sys.stderr)

    def confirm(self, msg, abort=False):
        """Get confirmation for an action."""
        return False

    def start_progress(self, name, total, **kwargs):
        """Start a new tqdm progressbar."""
        if name in self._progress_bars:
            raise ValueError("Name {} is already a registered progressbar.".format(name))

        if "type" not in kwargs:
            kwargs["type"] = "download"

        if kwargs["type"] not in self._progress_types:
            self._progress_bars[name] = None
        elif kwargs["type"] == "download":
            self._progress_bars[name] = tqdm(
                total=total,
                unit="iB",
                unit_scale=True,
                desc=name,
                leave=False,
                bar_format="{desc:.32}: {percentage:3.0f}%|{bar}{r_bar}",
            )

    def update_progress(self, name, amount):
        """Update a progressbar."""
        if name not in self._progress_bars or not self._progress_bars[name]:
            return

        self._progress_bars[name].update(amount)

    def finalize_progress(self, name):
        """End a progressbar."""
        if name not in self._progress_bars or not self._progress_bars[name]:
            return

        self._progress_bars[name].close()
        del self._progress_bars[name]


class ClickCallback(StandardOutput):
    """CommunicationCallback implementation for ``click`` messages."""

    INFO = click.style("Info: ", bold=True, fg="blue")
    WARNING = click.style("Warning: ", bold=True, fg="yellow")
    ERROR = click.style("Error: ", bold=True, fg="red")

    def echo(self, msg):
        """Write a message."""
        click.echo(msg)

    def info(self, msg):
        """Write an info message."""
        click.echo(self.INFO + msg)

    def warn(self, msg):
        """Write a warning message."""
        click.echo(self.WARNING + msg)

    def error(self, msg):
        """Write an error message."""
        click.echo(self.ERROR + msg)

    def confirm(self, msg, abort=False):
        """Get confirmation for an action using a prompt."""
        return click.confirm(msg, abort=abort)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration models V7."""

import os

from marshmallow import EXCLUDE

from renku.core.models import jsonld
from renku.core.models.calamus import fields, renku, schema

from .v3 import Base, DatasetFileSchemaV3, DatasetSchemaV3


class Dataset(Base):
    """Dataset migration model."""

    @classmethod
    def from_yaml(cls, path, client=None, commit=None):
        """Read content from YAML file."""
        data = jsonld.read_yaml(path)
        self = DatasetSchemaV7(client=client, commit=commit).load(data)
        self._metadata_path = path
        return self

    def to_yaml(self, path=None):
        """Write content to a YAML file."""
        from renku.core.management import LocalClient

        data = DatasetSchemaV7().dump(self)
        path = path or self._metadata_path or os.path.join(self.path, LocalClient.METADATA)
        jsonld.write_yaml(path=path, data=data)


class DatasetFileSchemaV7(DatasetFileSchemaV3):
    """DatasetFile schema."""

    based_on = fields.Nested(schema.isBasedOn, "DatasetFileSchemaV7", missing=None)
    source = fields.String(renku.source, missing=None)


class DatasetSchemaV7(DatasetSchemaV3):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Dataset
        model = Dataset
        unknown = EXCLUDE

    files = fields.Nested(schema.hasPart, DatasetFileSchemaV7, many=True)


def get_client_datasets(client):
    """Return Dataset migration models for a client."""
    paths = client.renku_datasets_path.rglob(client.METADATA)
    return [Dataset.from_yaml(path=path, client=client) for path in paths]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sidebar configuration."""

import os


def write_if_changed(fname, contents):
    """Overwrite file if contents have changed."""
    try:
        with open(fname, "r") as fp:
            old_contents = fp.read()
    except FileNotFoundError:
        old_contents = ""

    if old_contents != contents:
        with open(fname, "w") as fp:
            fp.write(contents)


def generate_sidebar(on_rtd, rtd_version, conf_api):
    """Generate the side bar."""
    do_gen = os.environ.get("SIDEBAR", None) == "1" or on_rtd

    lines = ["", ".. DO NOT MODIFY! THIS PAGE IS AUTOGENERATED!", ""]

    def toctree(name, max_depth=2, hidden=False, include_hidden=False):
        """Add a toctree entry."""
        lines.append(".. toctree::")

        if name:
            lines.append(f"    :caption: {name}")
        if hidden:
            lines.append("    :hidden:")
        if include_hidden:
            lines.append("    :includehidden:")
        lines.extend([f"    :maxdepth: {max_depth}", ""])

    def endl():
        """End a toctree."""
        lines.append("")

    def write(desc, link, repo):
        """Write a menu entry."""
        if conf_api == repo:
            args = desc, link
        elif not do_gen:
            return
        else:
            args = (
                desc,
                f"https://{repo}.readthedocs.io/en/{rtd_version}/{link}.html",
            )

        lines.append("    {} <{}>".format(*args))

    def write_external(desc, link):
        """Write a menu entry."""
        lines.append(f"    {desc} <{link}>")

    #
    # Specify the sidebar contents here
    #

    toctree(None, hidden=True)
    write("About Renku", "introduction/index", "renku")
    write("Getting Started", "getting_started", "renku")
    write("Using Renku", "user/index", "renku")
    write("Renku Platform", "platform", "renku")

    toctree(None, max_depth=3, hidden=True, include_hidden=False)
    write("Renku Client", "introduction", "renku-python")

    toctree(None, max_depth=1, hidden=True)
    write("Get in touch", "get_in_touch", "renku")
    write("Release Notes", "release_notes", "renku")
    write_external("License", "https://www.apache.org/licenses/LICENSE-2.0")
    endl()

    write_if_changed("_sidebar.rst.inc", "\n".join(lines))



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration models V8."""

import os
from pathlib import Path

from marshmallow import EXCLUDE, pre_dump

from renku.core.models import jsonld
from renku.core.models.calamus import Uri, fields, schema
from renku.core.models.entities import generate_file_id

from .v3 import CreatorMixinSchemaV3, DatasetTagSchemaV3, EntitySchemaV3, LanguageSchemaV3, PersonSchemaV3, UrlSchemaV3
from .v7 import Base, DatasetFileSchemaV7


class DatasetFile(Base):
    """DatasetFile migration model."""

    def __init__(self, **kwargs):
        """Initialize an instance."""
        super().__init__(**kwargs)

        if hasattr(self, "path") and (not self._id or self._id.startswith("_:")):
            hexsha = "UNCOMMITTED"
            if self.client and Path(self.path).exists():
                hexsha = self.client.find_previous_commit().hexsha

            self._id = generate_file_id(client=self.client, hexsha=hexsha, path=self.path)


class Dataset(Base):
    """Dataset migration model."""

    @classmethod
    def from_yaml(cls, path, client=None, commit=None):
        """Read content from YAML file."""
        data = jsonld.read_yaml(path)
        self = DatasetSchemaV8(client=client, commit=commit, flattened=True).load(data)
        self._metadata_path = path
        return self

    def to_yaml(self, path=None):
        """Write content to a YAML file."""
        from renku.core.management import LocalClient

        for file_ in self.files:
            file_._project = self._project

        data = DatasetSchemaV8(flattened=True).dump(self)
        path = path or self._metadata_path or os.path.join(self.path, LocalClient.METADATA)
        jsonld.write_yaml(path=path, data=data)


class DatasetFileSchemaV8(DatasetFileSchemaV7):
    """DatasetFile schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.DigitalDocument
        model = DatasetFile
        unknown = EXCLUDE


class DatasetSchemaV8(CreatorMixinSchemaV3, EntitySchemaV3):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Dataset
        model = Dataset
        unknown = EXCLUDE

    creators = fields.Nested(schema.creator, PersonSchemaV3, many=True)
    date_created = fields.DateTime(schema.dateCreated, missing=None)
    date_published = fields.DateTime(schema.datePublished, missing=None)
    description = fields.String(schema.description, missing=None)
    files = fields.Nested(schema.hasPart, DatasetFileSchemaV8, many=True)
    identifier = fields.String(schema.identifier)
    in_language = fields.Nested(schema.inLanguage, LanguageSchemaV3, missing=None)
    keywords = fields.List(schema.keywords, fields.String())
    license = Uri(schema.license, missing=None, allow_none=True)
    name = fields.String(schema.alternateName, missing=None)
    same_as = fields.Nested(schema.sameAs, UrlSchemaV3, missing=None)
    tags = fields.Nested(schema.subjectOf, DatasetTagSchemaV3, many=True)
    title = fields.String(schema.name)
    url = fields.String(schema.url)
    version = fields.String(schema.version, missing=None)

    @pre_dump
    def fix_license(self, data, **kwargs):
        """Fix license to be a string."""
        if isinstance(data.license, dict):
            data.license = data.license.get("http://schema.org/url", "")

        return data


def get_client_datasets(client):
    """Return Dataset migration models for a client."""
    paths = client.renku_datasets_path.rglob(client.METADATA)
    return [Dataset.from_yaml(path=path, client=client) for path in paths]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset tests."""
import copy
import datetime
import os
import shutil
import stat
from pathlib import Path

import pytest
from git import Repo

from renku.core import errors
from renku.core.commands.dataset import add_file, create_dataset, file_unlink, list_datasets, list_files
from renku.core.errors import OperationError, ParameterError
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.models.datasets import Dataset
from renku.core.models.provenance.agents import Person
from renku.core.utils.contexts import chdir
from tests.utils import assert_dataset_is_mutated, raises


@pytest.mark.parametrize(
    "scheme, path, overwrite, error",
    [
        ("", "temp", False, None),
        ("file://", "temp", True, None),
        ("", "tempp", False, errors.ParameterError),
        ("http://", "example.com/file1", False, None),
        ("https://", "example.com/file1", True, None),
        ("bla://", "file", False, errors.UrlSchemeNotSupported),
    ],
)
def test_data_add(scheme, path, overwrite, error, client, directory_tree, dataset_responses):
    """Test data import."""
    with raises(error):
        if path == "temp":
            path = str(directory_tree / "file1")

        with client.with_dataset("dataset", create=True) as d:
            d.creators = [Person(name="me", email="me@example.com", id="me_id")]

            client.add_data_to_dataset(d, ["{}{}".format(scheme, path)], overwrite=overwrite)

        target_path = os.path.join(DATA_DIR, "dataset", "file1")

        with open(target_path) as f:
            assert f.read() == "123"

        assert d.find_file(target_path)

        # check that the imported file is read-only
        assert not os.access(target_path, stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        # check the linking
        if scheme in ("", "file://"):
            shutil.rmtree("./data/dataset")
            with client.with_dataset("dataset") as d:
                d.creators = [Person(name="me", email="me@example.com", id="me_id")]
                client.add_data_to_dataset(d, ["{}{}".format(scheme, path)], overwrite=True)
            assert os.path.exists(target_path)


def test_data_add_recursive(directory_tree, client):
    """Test recursive data imports."""
    with client.with_dataset("dataset", create=True) as dataset:
        dataset.creators = [Person(name="me", email="me@example.com", id="me_id")]
        client.add_data_to_dataset(dataset, [str(directory_tree / "dir1")])

        assert os.path.basename(os.path.dirname(dataset.files[0].path)) == "dir1"


def test_git_repo_import(client_with_datasets, data_repository):
    """Test adding data from a local git repository."""
    dataset = client_with_datasets.load_dataset("dataset-1")
    client_with_datasets.add_data_to_dataset(dataset, [os.path.join(os.path.dirname(data_repository.git_dir), "dir1")])
    path2 = os.path.join(DATA_DIR, "dataset-1", "dir1", "file2")
    path3 = os.path.join(DATA_DIR, "dataset-1", "dir1", "file3")

    assert os.stat(path2)
    assert os.stat(path3)
    assert {path2, path3} == {f.path for f in dataset.files}


@pytest.mark.parametrize(
    "creators",
    [
        [Person(name="me", email="me@example.com")],
        [{"http://schema.org/name": "me", "http://schema.org/email": "me@example.com",}],
    ],
)
def test_creator_parse(creators):
    """Test that different options for specifying creators work."""
    dataset = Dataset(name="dataset", creators=creators)
    creator = Person(name="me", email="me@example.com")
    assert creator in dataset.creators

    # email check
    with pytest.raises(ValueError):
        Person(name="me", email="meexample.com")

    # creators must be a set or list of dicts or Person
    with pytest.raises(ValueError):
        Dataset(name="dataset", creators=["name"])


def test_creators_with_same_email(tmp_path):
    """Test creators with different names and same email address."""
    creators = [Person(name="me", email="me@example.com"), Person(name="me2", email="me@example.com")]
    dataset = Dataset(name="dataset", creators=creators)
    path = tmp_path / "dataset.yml"
    dataset.to_yaml(path=path)

    dataset = Dataset.from_yaml(path)
    assert 1 == len(dataset.creators)
    assert dataset.creators[0].name in ["me", "me2"]


def test_dataset_serialization(client_with_datasets):
    """Test dataset (de)serialization."""
    dataset = client_with_datasets.load_dataset("dataset-1")

    def read_value(key):
        return dataset_metadata.get(key)[0].get("@value")

    flattened_metadata = dataset.as_jsonld()
    dataset = Dataset.from_jsonld(flattened_metadata)

    # assert that all attributes found in metadata are set in the instance
    assert dataset.date_created
    assert dataset.creators
    assert dataset.identifier
    assert dataset.title
    assert dataset.path
    assert dataset._project

    dataset_metadata = [m for m in flattened_metadata if "Dataset" in str(m["@type"])][0]

    # check values
    assert str(dataset.date_created.isoformat()) == read_value("http://schema.org/dateCreated")
    assert dataset.identifier == read_value("http://schema.org/identifier")
    assert dataset.title == read_value("http://schema.org/name")
    assert dataset.path == read_value("http://www.w3.org/ns/prov#atLocation")


def test_create_dataset_custom_message(project):
    """Test create dataset custom message."""
    create_dataset("ds1", title="", description="", creators=[], commit_message="my awesome dataset")

    last_commit = Repo(".").head.commit
    assert "my awesome dataset" == last_commit.message


def test_list_datasets_default(project):
    """Test a default dataset listing."""
    create_dataset("ds1", title="", description="", creators=[], commit_message="my awesome dataset")

    datasets = list_datasets()

    assert isinstance(datasets, list)
    assert "ds1" in [dataset.title for dataset in datasets]


def test_list_files_default(project, tmpdir):
    """Test a default file listing."""
    create_dataset("ds1", title="", description="", creators=[], commit_message="my awesome dataset")
    data_file = tmpdir / Path("somefile")
    data_file.write_text("1,2,3", encoding="utf-8")

    add_file([str(data_file)], "ds1")
    files = list_files(datasets=["ds1"])

    assert isinstance(files, list)
    assert "somefile" in [file_.name for file_ in files]


def test_unlink_default(directory_tree, client):
    """Test unlink default behaviour."""
    with chdir(client.path):
        create_dataset("dataset")
        add_file([str(directory_tree / "dir1")], "dataset")

    with pytest.raises(ParameterError):
        file_unlink("dataset", (), ())


def test_mutate(client):
    """Test metadata change after dataset mutation."""
    dataset = Dataset(
        client=client,
        name="my-dataset",
        creators=[Person.from_string("John Doe <john.doe@mail.com>")],
        date_published=datetime.datetime.now(datetime.timezone.utc),
        same_as="http://some-url",
    )
    old_dataset = copy.deepcopy(dataset)

    dataset.mutate()

    mutator = Person.from_git(client.repo)
    assert_dataset_is_mutated(old=old_dataset, new=dataset, mutator=mutator)


def test_mutator_is_added_once(client):
    """Test mutator of a dataset is added only once to its creators list."""
    mutator = Person.from_git(client.repo)

    dataset = Dataset(
        client=client,
        name="my-dataset",
        creators=[mutator],
        date_published=datetime.datetime.now(datetime.timezone.utc),
        same_as="http://some-url",
    )
    old_dataset = copy.deepcopy(dataset)

    dataset.mutate()

    assert_dataset_is_mutated(old=old_dataset, new=dataset, mutator=mutator)
    assert 1 == len(dataset.creators)


def test_mutate_is_done_once():
    """Test dataset mutation can be done only once."""
    dataset = Dataset(name="my-dataset", creators=[])
    before_id = dataset._id

    dataset.mutate()
    after_id = dataset._id

    assert before_id != after_id

    dataset.mutate()
    assert after_id == dataset._id


def test_cannot_mutate_immutable_dataset():
    """Check immutable datasets cannot be modified."""
    dataset = Dataset(name="my-dataset", creators=[], immutable=True)

    with pytest.raises(OperationError):
        dataset.mutate()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Project API."""

import os

import pytest

from renku.api import Project


@pytest.mark.parametrize("sub_path", [".", "src", "src/notebooks"])
def test_get_project(client, sub_path):
    """Test getting Project context within a repository."""
    working_dir = client.path / sub_path
    working_dir.mkdir(exist_ok=True, parents=True)
    os.chdir(working_dir)

    with Project() as project:
        assert client.path == project.path


def test_get_project_multiple(client):
    """Test getting Project context multiple times within a repository."""
    with Project() as project_1:
        pass

    with Project() as project_2:
        pass

    assert project_1.path == project_2.path


def test_get_or_create_project(client):
    """Test getting Project context or creating one yileds similar results."""
    with Project() as project_1:
        pass

    project_2 = Project()

    assert project_1.path == project_2.path


def test_get_project_outside_a_renku_project(directory_tree):
    """Test creating a Project object in a non-renku directory."""
    os.chdir(directory_tree)

    with Project() as project:
        assert project.client is not None



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets remove controller."""
from renku.core.commands.dataset import dataset_remove
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadWithSyncOperation
from renku.service.serializers.datasets import DatasetRemoveRequest, DatasetRemoveResponseRPC
from renku.service.views import result_response


class DatasetsRemoveCtrl(ServiceCtrl, ReadWithSyncOperation):
    """Controller for datasets remove endpoint."""

    REQUEST_SERIALIZER = DatasetRemoveRequest()
    RESPONSE_SERIALIZER = DatasetRemoveResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a datasets remove controller."""
        self.ctx = DatasetsRemoveCtrl.REQUEST_SERIALIZER.load(request_data)

        if self.ctx.get("commit_message") is None:
            self.ctx["commit_message"] = "service: dataset remove {0}".format(self.ctx["name"])

        super(DatasetsRemoveCtrl, self).__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        return dataset_remove(self.ctx["name"], commit_message=self.ctx["commit_message"])

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        _, remote_branch = self.execute_and_sync()

        response = self.ctx
        response["remote_branch"] = remote_branch

        return result_response(DatasetsRemoveCtrl.RESPONSE_SERIALIZER, response)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tabular format helper functions."""

from collections import OrderedDict

from renku.core import errors
from renku.core.models.tabulate import tabulate as tabulate_


def tabulate(collection, columns, columns_mapping, columns_alignments=None):
    """Format collection with a tabular output."""
    if not columns:
        raise errors.ParameterError("Columns cannot be empty.")

    columns = [c.lower().strip() for c in columns.split(",") if c]

    headers, alignments = _make_headers(columns, columns_mapping, columns_alignments)

    # Sort based on the first requested field
    attr = list(headers.keys())[0]
    try:
        collection = sorted(collection, key=lambda d: getattr(d, attr))
    except TypeError:
        pass

    alignments = alignments if collection else None  # To avoid a tabulate bug

    return tabulate_(collection, headers=headers, colalign=alignments, disable_numparse=True)


def _make_headers(columns, columns_mapping, columns_alignments):
    columns_alignments = columns_alignments or {}
    headers = OrderedDict()
    alignments = []
    for column in columns:
        if column not in columns_mapping:
            raise errors.ParameterError(
                'Invalid column name: "{}".\nPossible values: {}'.format(column, ", ".join(columns_mapping))
            )
        name, display_name = columns_mapping.get(column)
        headers[name] = display_name

        alignment = columns_alignments.get(column, "left")
        alignments.append(alignment)

    return headers, alignments



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serialization tests for renku models."""

import datetime
from urllib.parse import urljoin

from renku.core.management.client import LocalClient
from renku.core.models.datasets import Dataset
from renku.core.utils.uuid import is_uuid


def test_dataset_deserialization(client_with_datasets):
    """Test Dataset deserialization."""
    dataset = Dataset.from_yaml(client_with_datasets.get_dataset_path("dataset-1"), client=client_with_datasets)

    dataset_types = {
        "date_created": [datetime.datetime],
        "creators": [list],
        "description": [str, type(None)],
        "files": [list],
        "identifier": [str],
        "keywords": [list],
    }

    for attribute, type_ in dataset_types.items():
        assert type(dataset.__getattribute__(attribute)) in type_

    creator_types = {"email": str, "_id": str, "name": str, "affiliation": str}

    creator = client_with_datasets.load_dataset("dataset-1").creators[0]

    for attribute, type_ in creator_types.items():
        assert type(getattr(creator, attribute)) is type_


def test_dataset_files_empty_metadata(dataset_metadata):
    """Check parsing metadata of dataset files with empty filename."""
    dataset = Dataset.from_jsonld(dataset_metadata, client=LocalClient("."),)
    files = [file.filename for file in dataset.files if not file.filename]

    if files:
        assert None in files


def test_uuid_migration(dataset_metadata, client):
    """Test migration of id with UUID."""
    dataset = Dataset.from_jsonld(dataset_metadata, client=client)

    assert is_uuid(dataset.identifier)
    assert urljoin("https://localhost/datasets/", dataset.identifier) == dataset._id


def test_dataset_creator_email(dataset_metadata):
    """Check that creators without an email are assigned a blank node."""
    # modify the dataset metadata to change the creator
    dataset = Dataset.from_jsonld(dataset_metadata, client=LocalClient("."),)

    dataset.creators[0]._id = "mailto:None"
    dataset_broken = Dataset.from_jsonld(dataset.as_jsonld(), client=LocalClient("."))
    assert "mailto:None" not in dataset_broken.creators[0]._id


def test_calamus(client, dataset_metadata_before_calamus):
    """Check Calamus loads project correctly."""
    dataset = Dataset.from_jsonld(dataset_metadata_before_calamus, client=LocalClient("."))

    file_ = dataset.find_file("data/dataverse/external/data.txt")
    assert file_.external is True
    assert "file://../../../../tmp/data.txt" == file_.url

    file_ = dataset.find_file("data/dataverse/local/result.csv")
    assert file_.external is False
    assert "file://../../../../tmp/result.csv" == file_.url


def test_dataset_with_multiple_project_version(client_with_datasets):
    """Test deserialization of a dataset where contains different project versions."""
    max_version = "42000"

    # Change project version for a single file
    with client_with_datasets.with_dataset("dataset-2") as dataset:
        file_ = dataset.find_file(dataset.data_dir / "file1")
        file_._project.version = max_version

    dataset = client_with_datasets.load_dataset("dataset-2")

    assert {max_version} == {f._project.version for f in dataset.files}



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for renku service controllers."""
import shutil

from renku.core.commands.clone import project_clone
from renku.core.errors import RenkuException
from renku.service.logger import service_log
from renku.service.utils import make_project_path
from renku.service.views.decorators import requires_cache


@requires_cache
def user_project_clone(cache, user_data, project_data):
    """Clones the project for a given user."""
    local_path = make_project_path(user_data, project_data)
    if local_path is None:
        raise RenkuException("project not found")

    user = cache.ensure_user(user_data)

    if local_path.exists():
        shutil.rmtree(local_path)

        for project in cache.get_projects(user):
            if project.abs_path == local_path:
                project.delete()

        if "project_id" in project_data:
            project_data.pop("project_id")

    repo, initialized = project_clone(
        project_data["url_with_auth"],
        local_path,
        depth=project_data["depth"] if project_data["depth"] != 0 else None,
        raise_git_except=True,
        config={"user.name": project_data["fullname"], "user.email": project_data["email"],},
        checkout_rev=project_data["ref"],
    )
    project_data["initialized"] = initialized

    service_log.debug(f"project successfully cloned: {repo}")
    service_log.debug(f"project folder exists: {local_path.exists()}")

    project = cache.make_project(user, project_data)
    return project



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates read manifest controller tests."""
import pytest
from git import GitCommandError
from marshmallow import ValidationError


def test_template_read_manifest_ctrl(ctrl_init, svc_client_with_templates, mocker):
    """Test template read manifest controller."""
    from renku.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl

    cache, user_data = ctrl_init
    svc_client, headers, template_params = svc_client_with_templates

    ctrl = TemplatesReadManifestCtrl(cache, user_data, template_params)
    ctrl_mock = mocker.patch.object(
        ctrl, "template_manifest", return_value=[{"name": "test", "folder": ".", "description": "desc"}]
    )

    response = ctrl.to_response()
    assert {"result": {"templates": ctrl_mock.return_value}} == response.json


@pytest.mark.parametrize("git_url", ["https://github.com", "https://github.com/SwissDataScienceCenter",])
def test_validation_exc_template_read_manifest_ctrl(git_url, ctrl_init, svc_client_with_templates, mocker):
    """Test validation exception on template read manifest controller."""
    from renku.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl

    cache, user_data = ctrl_init
    svc_client, headers, template_params = svc_client_with_templates
    template_params["url"] = git_url

    with pytest.raises(ValidationError):
        TemplatesReadManifestCtrl(cache, user_data, template_params)


@pytest.mark.service
@pytest.mark.integration
@pytest.mark.parametrize("git_url", ["https://test.com/test2/test3", "https://www.test.com/test2/test3"])
def test_found_exc_template_read_manifest_ctrl(git_url, ctrl_init, svc_client_with_templates, mocker):
    """Test git command exception on template read manifest controller."""
    from renku.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl

    cache, user_data = ctrl_init
    svc_client, headers, template_params = svc_client_with_templates
    template_params["url"] = git_url

    ctrl = TemplatesReadManifestCtrl(cache, user_data, template_params)

    with pytest.raises(GitCommandError):
        ctrl.to_response()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a Git commit."""

import os
import urllib
import weakref
from collections import OrderedDict
from pathlib import Path, posixpath

import attr
from git import NULL_TREE
from marshmallow import EXCLUDE

from renku.core.models import jsonld
from renku.core.models.calamus import Nested, fields, oa, prov, rdfs, renku, wfprov
from renku.core.models.cwl.annotation import AnnotationSchema
from renku.core.models.entities import (
    Collection,
    CollectionSchema,
    CommitMixin,
    CommitMixinSchema,
    Entity,
    EntitySchema,
)
from renku.core.models.refs import LinkReference
from renku.core.models.workflow.run import Run
from renku.core.utils.scm import git_unicode_unescape

from ..workflow.parameters import RunParameter, RunParameterSchema
from .agents import Person, PersonSchema, SoftwareAgentSchema, renku_agent
from .qualified import Association, AssociationSchema, Generation, GenerationSchema, Usage, UsageSchema


def _nodes(output, parent=None):
    """Yield nodes from entities."""
    # NOTE refactor so all outputs behave the same
    entity = getattr(output, "entity", output)

    if isinstance(entity, Collection):
        for member in entity.members:
            if parent is not None:
                member = attr.evolve(member, parent=parent)

            if entity.client:
                _set_entity_client_commit(member, entity.client, None)
            if isinstance(output, Generation):
                child = Generation(
                    activity=output.activity, entity=member, role=entity.role if hasattr(entity, "role") else None
                )
            elif isinstance(output, Usage):
                child = Usage(
                    activity=output.activity, entity=member, role=entity.role if hasattr(entity, "role") else None
                )
            else:
                child = member
            yield from _nodes(child)

    yield output


def _set_entity_client_commit(entity, client, commit):
    """Set the client and commit of an entity."""
    if client and not entity.client:
        entity.client = client

    if not entity.commit:
        revision = "UNCOMMITTED"
        if entity._label:
            revision = entity._label.rsplit("@")[1]
        if revision == "UNCOMMITTED":
            commit = commit
        elif client:
            commit = client.repo.commit(revision)
        entity.commit = commit


@attr.s(cmp=False,)
class Activity(CommitMixin):
    """Represent an activity in the repository."""

    _id = attr.ib(default=None, kw_only=True)
    _message = attr.ib(kw_only=True)
    _was_informed_by = attr.ib(kw_only=True,)

    part_of = attr.ib(default=None, kw_only=True)

    _collections = attr.ib(default=attr.Factory(OrderedDict), init=False, kw_only=True)
    generated = attr.ib(kw_only=True, default=None)

    invalidated = attr.ib(kw_only=True, default=None)

    influenced = attr.ib(kw_only=True)

    started_at_time = attr.ib(kw_only=True)

    ended_at_time = attr.ib(kw_only=True)

    agents = attr.ib(kw_only=True)

    _metadata_path = attr.ib(default=None, init=False)

    def default_generated(self):
        """Create default ``generated``."""
        generated = []

        for path in self.get_output_paths():
            entity = self._get_activity_entity(path)

            generated.append(Generation(activity=self, entity=entity, role=None))
        return generated

    def get_output_paths(self):
        """Gets all output paths generated by this run."""
        index = set()

        commit = self.commit

        if not self.commit:
            if not self.client:
                return index
            commit = self.client.repo.head.commit

        for file_ in commit.diff(commit.parents or NULL_TREE):
            # ignore deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type == "A":
                continue

            path_ = Path(git_unicode_unescape(file_.a_path))

            is_dataset = any(
                [
                    path_.resolve() == (self.client.path / f.path).resolve()
                    for d in self.client.datasets.values()
                    for f in d.files
                ]
            )
            not_refs = LinkReference.REFS not in str(path_)
            does_not_exists = not path_.exists()

            if all([is_dataset, not_refs, does_not_exists]):
                dataset = next(
                    d
                    for d in self.client.datasets.values()
                    for f in d.files
                    if path_.resolve() == (self.client.path / f.path).resolve()
                )
                path_ = self.client.path / dataset.path / self.client.METADATA

            index.add(str(path_))

        return index

    def _get_activity_entity(self, path, deleted=False):
        """Gets the entity associated with this Activity and path."""
        client, commit, path = self.client.resolve_in_submodules(self.commit, path,)
        output_path = client.path / path
        parents = list(output_path.relative_to(client.path).parents)

        collection = None
        members = []
        for parent in reversed(parents[:-1]):
            if str(parent) in self._collections:
                collection = self._collections[str(parent)]
            else:
                collection = Collection(client=client, commit=commit, path=str(parent), members=[], parent=collection,)
                members.append(collection)
                self._collections[str(parent)] = collection

            members = collection.members

        entity_cls = Entity
        if (self.client.path / path).is_dir():
            entity_cls = Collection

        # TODO: use a factory method to generate the entity
        if str(path).startswith(os.path.join(client.renku_home, client.DATASETS)) and not deleted:
            entity = client.load_dataset_from_path(path, commit=commit)
        else:
            entity = entity_cls(commit=commit, client=client, path=str(path), parent=collection,)

        if collection:
            collection.members.append(entity)

        return entity

    def default_invalidated(self):
        """Entities invalidated by this Action."""
        results = []
        for path in self.removed_paths:
            entity = self._get_activity_entity(path, deleted=True)

            results.append(entity)
        return results

    @influenced.default
    def default_influenced(self):
        """Calculate default values."""
        return list(self._collections.values())

    @property
    def parents(self):
        """Return parent commits."""
        if self.commit:
            return list(self.commit.parents)

    @property
    def removed_paths(self):
        """Return all paths removed in the commit."""
        index = set()
        if not self.commit:
            return index

        for file_ in self.commit.diff(self.commit.parents or NULL_TREE):
            # only process deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type != "A":
                continue
            path_ = Path(file_.a_path)

            index.add(str(path_))

        return index

    @property
    def paths(self):
        """Return all paths in the commit."""
        index = set()

        for file_ in self.commit.diff(self.commit.parents or NULL_TREE):
            # ignore deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type == "A":
                continue
            path_ = Path(git_unicode_unescape(file_.a_path))

            is_dataset = any(
                [
                    path_.resolve() == (self.client.path / f.path).resolve()
                    for d in self.client.datasets.values()
                    for f in d.files
                ]
            )
            not_refs = LinkReference.REFS not in str(path_)
            does_not_exists = not (path_.exists() or (path_.is_symlink() and os.path.lexists(path_)))

            if all([is_dataset, not_refs, does_not_exists]):
                dataset = next(
                    d
                    for d in self.client.datasets
                    for f in d.files
                    if path_.resolve() == (self.client.path / f.path).resolve()
                )
                path_ = self.client.path / dataset.path / self.client.METADATA

            index.add(str(path_))

        return index

    @classmethod
    def generate_id(cls, commitsha):
        """Calculate action ID."""
        host = "localhost"
        if hasattr(cls, "client"):
            host = cls.client.remote.get("host") or host
        host = os.environ.get("RENKU_DOMAIN") or host

        return urllib.parse.urljoin(
            "https://{host}".format(host=host),
            posixpath.join("/activities", "commit/{commit}".format(commit=commitsha)),
        )

    def default_id(self):
        """Configure calculated ID."""
        if self.commit:
            return self.generate_id(self.commit.hexsha)
        return self.generate_id("UNCOMMITTED")

    @_message.default
    def default_message(self):
        """Generate a default message."""
        if self.commit:
            return self.commit.message

    @_was_informed_by.default
    def default_was_informed_by(self):
        """List parent actions."""
        if self.commit:
            return [self.generate_id(parent) for parent in self.commit.parents]

    @started_at_time.default
    def default_started_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.authored_datetime

    @ended_at_time.default
    def default_ended_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.committed_datetime

    @agents.default
    def default_agents(self):
        """Set person agent to be the author of the commit."""
        if self.commit:
            return [Person.from_commit(self.commit), renku_agent]
        return [renku_agent]

    @property
    def nodes(self):
        """Return topologically sorted nodes."""
        collections = OrderedDict()

        def _parents(node):
            if node.parent:
                yield from _parents(node.parent)
                yield node.parent

        for output in self.generated:
            for parent in _parents(output.entity):
                collections[parent.path] = parent

            yield from _nodes(output)

        for removed in self.invalidated:
            for parent in _parents(removed):
                collections[parent.path] = parent

            yield from _nodes(removed)

        yield from reversed(collections.values())

    def __attrs_post_init__(self):
        """Sets ``generated`` and ``invalidated`` default values if needed."""
        super().__attrs_post_init__()
        if not self._id:
            self._id = self.default_id()
        if not self.generated:
            self.generated = self.default_generated()

        for g in self.generated:
            _set_entity_client_commit(g.entity, self.client, self.commit)

        if not self.invalidated:
            self.invalidated = self.default_invalidated()

        if self.generated:
            for g in self.generated:
                g._activity = weakref.ref(self)

    @classmethod
    def from_yaml(cls, path, client=None, commit=None):
        """Return an instance from a YAML file."""
        data = jsonld.read_yaml(path)

        self = cls.from_jsonld(data=data, client=client, commit=commit)
        self._metadata_path = path

        return self

    def to_yaml(self, path=None):
        """Write an instance to the referenced YAML file."""
        self._metadata_path = path or self._metadata_path
        data = ActivitySchema(flattened=True).dump(self)
        jsonld.write_yaml(path=self._metadata_path, data=data)

    @classmethod
    def from_jsonld(cls, data, client=None, commit=None):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, list):
            raise ValueError(data)

        schema = ActivitySchema

        if any(str(wfprov.WorkflowRun) in d["@type"] for d in data):
            schema = WorkflowRunSchema
        elif any(str(wfprov.ProcessRun) in d["@type"] for d in data):
            schema = ProcessRunSchema

        return schema(client=client, commit=commit, flattened=True).load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return ActivitySchema(flattened=True).dump(self)


@attr.s(cmp=False,)
class ProcessRun(Activity):
    """A process run is a particular execution of a Process description."""

    __association_cls__ = Run

    generated = attr.ib(kw_only=True, default=None)

    association = attr.ib(default=None, kw_only=True)

    annotations = attr.ib(kw_only=True, default=None)

    qualified_usage = attr.ib(kw_only=True, default=None)

    run_parameter = attr.ib(kw_only=True, default=None)

    def __attrs_post_init__(self):
        """Calculate properties."""
        super().__attrs_post_init__()
        commit_not_set = not self.commit or self.commit.hexsha in self._id
        if commit_not_set and self.client and Path(self.path).exists():
            self.commit = self.client.find_previous_commit(self.path)

        if not self.annotations:
            self.annotations = self.plugin_annotations()

        if self.association:
            self.association.plan._activity = weakref.ref(self)
            plan = self.association.plan
            if not plan.commit:
                if self.client:
                    plan.client = self.client
                if self.commit:
                    plan.commit = self.commit

                if plan.inputs:
                    for i in plan.inputs:
                        _set_entity_client_commit(i.consumes, self.client, self.commit)
                if plan.outputs:
                    for o in plan.outputs:
                        _set_entity_client_commit(o.produces, self.client, self.commit)

        if self.qualified_usage and self.client and self.commit:
            usages = []
            revision = "{0}".format(self.commit)
            for usage in self.qualified_usage:
                if not usage.commit and "@UNCOMMITTED" in usage._label:
                    usages.append(
                        Usage.from_revision(
                            client=self.client, path=usage.path, role=usage.role, revision=revision, id=usage._id,
                        )
                    )
                else:
                    if not usage.client:
                        usage.entity.set_client(self.client)
                    if not usage.commit:
                        revision = usage._label.rsplit("@")[1]
                        usage.entity.commit = self.client.repo.commit(revision)

                    usages.append(usage)
            self.qualified_usage = usages

    def default_generated(self):
        """Create default ``generated``."""
        generated = []

        if not self.association or not self.association.plan:
            return generated

        for output in self.association.plan.outputs:
            entity = Entity.from_revision(
                self.client, output.produces.path, revision=self.commit, parent=output.produces.parent
            )

            generation = Generation(activity=self, role=output.sanitized_id, entity=entity)
            generated.append(generation)
        return generated

    def add_annotations(self, annotations):
        """Adds annotations from an external tool."""
        self.annotations.extend(annotations)

    def plugin_annotations(self):
        """Adds ``Annotation``s from plugins to a ``ProcessRun``."""
        from renku.core.plugins.pluginmanager import get_plugin_manager

        pm = get_plugin_manager()

        results = pm.hook.process_run_annotations(run=self)
        return [a for r in results for a in r]

    @classmethod
    def from_run(cls, run, client, path, commit=None, subprocess_index=None, update_commits=False):
        """Convert a ``Run`` to a ``ProcessRun``."""
        from .agents import SoftwareAgent

        if not commit:
            commit = client.repo.head.commit

        usages = []

        id_ = ProcessRun.generate_id(commit)

        if subprocess_index is not None:
            id_ = f"{id_}/steps/step_{subprocess_index}"

        for input_ in run.inputs:
            usage_id = f"{id_}/{input_.sanitized_id}"
            input_path = input_.consumes.path
            entity = input_.consumes
            if update_commits:
                revision = client.find_previous_commit(input_path, revision=commit.hexsha)
                entity = Entity.from_revision(client, input_path, revision)

            dependency = Usage(entity=entity, role=input_.sanitized_id, id=usage_id)

            usages.append(dependency)

        agent = SoftwareAgent.from_commit(commit)
        association = Association(agent=agent, id=id_ + "/association", plan=run)

        run_parameter = []

        for parameter in run.run_parameters:
            parameter_id = f"{id_}/{parameter.name}"
            run_parameter.append(RunParameter(name=parameter.name, value=parameter.value, id=parameter_id))

        process_run = cls(
            id=id_,
            qualified_usage=usages,
            association=association,
            client=client,
            commit=commit,
            path=path,
            run_parameter=run_parameter,
        )

        generated = []

        for output in run.outputs:
            entity = Entity.from_revision(client, output.produces.path, revision=commit, parent=output.produces.parent)

            generation = Generation(activity=process_run, role=output.sanitized_id, entity=entity)
            generated.append(generation)

        process_run.generated = generated

        process_run.plugin_annotations()
        return process_run

    @property
    def parents(self):
        """Return parent commits."""
        return [member.commit for usage in self.qualified_usage for member in usage.entity.entities] + super().parents

    @property
    def nodes(self):
        """Return topologically sorted nodes."""
        # Outputs go first
        yield from super().nodes

        # Activity itself
        yield self.association.plan

    def to_yaml(self, path=None):
        """Write an instance to the referenced YAML file."""
        self._metadata_path = path or self._metadata_path
        data = ProcessRunSchema(flattened=True).dump(self)
        jsonld.write_yaml(path=self._metadata_path, data=data)

    @classmethod
    def from_jsonld(cls, data, client=None, commit=None):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, list):
            raise ValueError(data)

        return ProcessRunSchema(client=client, commit=commit, flattened=True).load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return ProcessRunSchema(flattened=True).dump(self)


@attr.s(cmp=False,)
class WorkflowRun(ProcessRun):
    """A workflow run typically contains several subprocesses."""

    __association_cls__ = Run

    _processes = attr.ib(kw_only=True, default=attr.Factory(list))

    @property
    def subprocesses(self):
        """Subprocesses of this ``WorkflowRun``."""
        return {i: p for i, p in enumerate(self._processes)}

    @classmethod
    def from_run(cls, run, client, path, commit=None, update_commits=False):
        """Convert a ``Run`` to a ``WorkflowRun``."""
        from .agents import SoftwareAgent

        if not commit:
            commit = client.repo.head.commit

        processes = []
        generated = []

        for s in run.subprocesses:
            proc_run = ProcessRun.from_run(s.process, client, path, commit, s.index, update_commits)
            processes.append(proc_run)
            generated.extend(proc_run.generated)

        usages = []

        id_ = cls.generate_id(commit)
        input_index = 1
        for input_ in run.inputs:
            usage_id = f"{id_}/inputs/{input_index}"

            dependency = Usage.from_revision(
                client=client, path=input_.consumes.path, role=input_.sanitized_id, revision=commit, id=usage_id,
            )

            usages.append(dependency)
            input_index += 1

        agent = SoftwareAgent.from_commit(commit)
        association = Association(agent=agent, id=id_ + "/association", plan=run)

        all_generated = []

        # fix generations in folders
        for generation in generated:
            all_generated.append(generation)
            entity = generation.entity

            if not isinstance(entity, Collection) or not entity.commit:
                continue

            for e in entity.entities:
                if e.commit is not entity.commit or any(g.entity._id == e._id for g in all_generated):
                    continue

                all_generated.append(Generation(activity=generation.activity, entity=e, role=None))

        wf_run = WorkflowRun(
            id=id_,
            processes=processes,
            generated=all_generated,
            qualified_usage=usages,
            association=association,
            client=client,
            commit=commit,
            path=path,
        )
        return wf_run

    @property
    def nodes(self):
        """Yield all graph nodes."""
        for subprocess in reversed(self._processes):
            if subprocess.path is None:
                # skip nodes connecting directory to file
                continue

            for n in subprocess.nodes:
                # if self.client and not n.commit and isinstance(n, Entity):
                #     _set_entity_client_commit(n, self.client, self.commit)
                # n._activity = weakref.ref(subprocess)
                yield n
            yield subprocess.association.plan

    def __attrs_post_init__(self):
        """Attrs post initializations."""
        if not self._id:
            self._id = self.default_id()

        if not self._processes:
            self._processes = []
            for subprocess in self.association.plan.subprocesses:
                run = subprocess.process
                process_run = ProcessRun.from_run(
                    run=run,
                    client=self.client,
                    path=self.path,
                    commit=self.commit,
                    subprocess_index=subprocess.index,
                    update_commits=True,
                )

                self._processes.append(process_run)

        if self.client:
            for s in self._processes:
                s.client = self.client
                s.commit = self.commit
                s.__attrs_post_init__()
                s.part_of = self

        super().__attrs_post_init__()

    def to_yaml(self, path=None):
        """Write an instance to the referenced YAML file."""
        self._metadata_path = path or self._metadata_path
        data = WorkflowRunSchema(flattened=True).dump(self)
        jsonld.write_yaml(path=self._metadata_path, data=data)

    @classmethod
    def from_jsonld(cls, data, client=None, commit=None):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, list):
            raise ValueError(data)

        return WorkflowRunSchema(client=client, commit=commit, flattened=True).load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return WorkflowRunSchema(flattened=True).dump(self)


class ActivitySchema(CommitMixinSchema):
    """Activity schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Activity
        model = Activity
        unknown = EXCLUDE

    _message = fields.String(rdfs.comment, init_name="message", missing=None)
    _was_informed_by = fields.List(prov.wasInformedBy, fields.IRI(), init_name="was_informed_by")
    generated = Nested(prov.activity, GenerationSchema, reverse=True, many=True, missing=None)
    invalidated = Nested(prov.wasInvalidatedBy, [EntitySchema, CollectionSchema], reverse=True, many=True, missing=None)
    influenced = Nested(prov.influenced, CollectionSchema, many=True)
    started_at_time = fields.DateTime(prov.startedAtTime, add_value_types=True)
    ended_at_time = fields.DateTime(prov.endedAtTime, add_value_types=True)
    agents = Nested(prov.wasAssociatedWith, [PersonSchema, SoftwareAgentSchema], many=True)


class ProcessRunSchema(ActivitySchema):
    """ProcessRun schema."""

    class Meta:
        """Meta class."""

        rdf_type = wfprov.ProcessRun
        model = ProcessRun
        unknown = EXCLUDE

    association = Nested(prov.qualifiedAssociation, AssociationSchema)
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    qualified_usage = Nested(prov.qualifiedUsage, UsageSchema, many=True)
    run_parameter = Nested(renku.hasRunParameter, RunParameterSchema, many=True)


class WorkflowRunSchema(ProcessRunSchema):
    """WorkflowRun schema."""

    class Meta:
        """Meta class."""

        rdf_type = wfprov.WorkflowRun
        model = WorkflowRun
        unknown = EXCLUDE

    _processes = Nested(wfprov.wasPartOfWorkflowRun, ProcessRunSchema, reverse=True, many=True, init_name="processes")



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``update`` command."""

from pathlib import Path

import git

from renku.cli import cli
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.models.entities import Collection


def update_and_commit(data, file_, repo):
    """Update source.txt."""
    with file_.open("w") as fp:
        fp.write(data)

    repo.git.add(file_)
    repo.index.commit("Updated source.txt")


def test_update(runner, project, renku_cli, no_lfs_warning):
    """Test automatic file update."""
    from renku.core.utils.shacl import validate_graph

    cwd = Path(project)
    data = cwd / DATA_DIR
    data.mkdir(exist_ok=True, parents=True)
    source = cwd / "source.txt"
    output = data / "result.txt"

    repo = git.Repo(project)

    update_and_commit("1", source, repo)

    exit_code, run = renku_cli("run", "wc", "-c", stdin=source, stdout=output)
    assert 0 == exit_code
    assert 0 == len(run.subprocesses)
    previous_run_id = run._id

    with output.open("r") as f:
        assert f.read().strip() == "1"

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    update_and_commit("12", source, repo)

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code

    exit_code, run = renku_cli("update", "--all")
    assert 0 == exit_code
    assert 0 == len(run.subprocesses)
    assert previous_run_id == run._id
    previous_run_id = run._id

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    with output.open("r") as f:
        assert f.read().strip() == "2"

    # Source has been updated but output is unchanged.
    update_and_commit("34", source, repo)

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code

    exit_code, run = renku_cli("update", "--all")
    assert 0 == exit_code
    assert 0 == len(run.subprocesses)
    assert previous_run_id == run._id

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    with output.open("r") as f:
        assert f.read().strip() == "2"

    from renku.cli.log import FORMATS

    for output_format in FORMATS:
        # Make sure the log contains the original parent.
        result = runner.invoke(cli, ["log", "--format", output_format], catch_exceptions=False,)
        assert 0 == result.exit_code, result.output
        assert source.name in result.output, output_format

        if output_format == "nt":
            r, _, t = validate_graph(result.output, format="nt")
            assert r is True, t


def test_update_multiple_steps(runner, project, renku_cli, no_lfs_warning):
    """Test automatic file update."""
    cwd = Path(project)
    data = cwd / "data"
    data.mkdir(exist_ok=True, parents=True)
    source = cwd / "source.txt"
    intermediate = cwd / "intermediate.txt"
    output = cwd / "result.txt"

    repo = git.Repo(project)

    update_and_commit("1", source, repo)

    exit_code, run = renku_cli("run", "cp", str(source), str(intermediate))
    assert 0 == exit_code
    assert 0 == len(run.subprocesses)

    with intermediate.open("r") as f:
        assert f.read().strip() == "1"

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    exit_code, run = renku_cli("run", "cp", str(intermediate), str(output))
    assert 0 == exit_code
    assert 0 == len(run.subprocesses)

    with output.open("r") as f:
        assert f.read().strip() == "1"

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    update_and_commit("2", source, repo)

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code

    exit_code, run = renku_cli("update", "--all")
    assert 0 == exit_code
    assert 2 == len(run.subprocesses)

    result = runner.invoke(cli, ["log"], catch_exceptions=False)
    assert "(part of" in result.output, result.output

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    with output.open("r") as f:
        assert f.read().strip() == "2"


def test_workflow_without_outputs(runner, project, run):
    """Test workflow without outputs."""
    repo = git.Repo(project)
    cwd = Path(project)
    input_ = cwd / "input.txt"
    with input_.open("w") as f:
        f.write("first")

    repo.git.add("--all")
    repo.index.commit("Created input.txt")

    cmd = ["run", "cat", "--no-output", input_.name]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    cmd = ["status", "--no-output"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    with input_.open("w") as f:
        f.write("second")

    repo.git.add("--all")
    repo.index.commit("Updated input.txt")

    cmd = ["status", "--no-output"]
    result = runner.invoke(cli, cmd)
    assert 1 == result.exit_code
    assert 0 == run(args=("update", "--no-output", "--all"))

    cmd = ["status", "--no-output"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code


def test_siblings_update(runner, project, run, no_lfs_warning):
    """Test detection of siblings during update."""
    cwd = Path(project)
    parent = cwd / "parent.txt"
    brother = cwd / "brother.txt"
    sister = cwd / "sister.txt"
    siblings = {brother, sister}

    repo = git.Repo(project)

    def update_source(data):
        """Update parent.txt."""
        with parent.open("w") as fp:
            fp.write(data)

        repo.git.add("--all")
        repo.index.commit("Updated parent.txt")

    update_source("1")

    # The output files do not exist.
    assert not any(sibling.exists() for sibling in siblings)

    cmd = ["run", "tee", "brother.txt"]
    assert 0 == run(args=cmd, stdin=parent, stdout=sister)

    # The output file is copied from the source.
    for sibling in siblings:
        with sibling.open("r") as f:
            assert f.read().strip() == "1", sibling

    update_source("2")

    # Siblings must be updated together.
    for sibling in siblings:
        assert 1 == run(args=("update", sibling.name))

    # Update brother and check the sister has not been changed.
    assert 0 == run(args=("update", "--with-siblings", brother.name))

    for sibling in siblings:
        with sibling.open("r") as f:
            assert f.read().strip() == "2", sibling

    update_source("3")

    # Siblings kept together even when one is removed.
    repo.index.remove([brother.name], working_tree=True)
    repo.index.commit("Brother removed")

    assert not brother.exists()

    # Update should find also missing siblings.
    assert 1 == run(args=("update", "--all",))
    assert 0 == run(args=("update", "--with-siblings", "--all"))

    for sibling in siblings:
        with sibling.open("r") as f:
            assert f.read().strip() == "3", sibling


def test_siblings_in_output_directory(runner, project, run):
    """Files in output directory are linked or removed after update."""
    repo = git.Repo(project)
    cwd = Path(project)
    source = cwd / "source.txt"
    output = cwd / "output"

    files = [
        ("first", "1"),
        ("second", "2"),
        ("third", "3"),
    ]

    def write_source():
        """Write source from files."""
        with source.open("w") as fp:
            fp.write("\n".join(" ".join(line) for line in files) + "\n")

        repo.git.add("--all")
        repo.index.commit("Update source.txt")

    def check_files():
        """Check file content."""
        assert len(files) == len(list(output.rglob("*")))

        for name, content in files:
            with (output / name).open() as fp:
                assert content == fp.read().strip(), name

    write_source()

    script = 'mkdir -p "$0"; ' "cat - | while read -r name content; do " 'echo "$content" > "$0/$name"; done'
    base_sh = ["sh", "-c", script, "output"]

    assert not output.exists()
    assert 0 == run(args=["run"] + base_sh, stdin=source)
    assert output.exists()
    check_files()

    files = [
        ("first", "11"),
        ("third", "3"),
        ("fourth", "4"),
    ]
    write_source()
    assert 0 == run(args=["update", "output"])
    check_files()


def test_relative_path_for_directory_input(client, run, renku_cli):
    """Test having a directory input generates relative path in CWL."""
    (client.path / DATA_DIR / "file1").write_text("file1")
    client.repo.git.add("--all")
    client.repo.index.commit("Add file")

    assert 0 == run(args=["run", "ls", DATA_DIR], stdout="ls.data")

    (client.path / DATA_DIR / "file2").write_text("file2")
    client.repo.git.add("--all")
    client.repo.index.commit("Add one more file")

    exit_code, cwl = renku_cli("update", "--all")
    assert 0 == exit_code
    assert 1 == len(cwl.inputs)
    assert isinstance(cwl.inputs[0].consumes, Collection)
    assert "data" == cwl.inputs[0].consumes.path


def test_update_no_args(runner, project, renku_cli, no_lfs_warning):
    """Test calling update with no args does nothing."""
    cwd = Path(project)
    data = cwd / DATA_DIR
    data.mkdir(exist_ok=True, parents=True)
    source = cwd / "source.txt"
    output = data / "result.txt"

    repo = git.Repo(project)

    update_and_commit("1", source, repo)

    exit_code, run = renku_cli("run", "wc", "-c", stdin=source, stdout=output)
    assert 0 == exit_code

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code

    update_and_commit("12", source, repo)

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code

    before_commit = repo.head.commit

    exit_code, run = renku_cli("update")
    assert 0 == exit_code

    assert before_commit == repo.head.commit



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for renku service controllers."""
import collections


def squash(data, parent_key="", sep="."):
    """Squash deeply nested dictionary."""
    items = []
    for k, v in data.items():
        k = str(k)
        if k[0] == "_":
            k = k[1:]

        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(squash(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))

    return dict(items)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets import controller."""
import os

from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadWithSyncOperation
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.datasets import dataset_import
from renku.service.jobs.queues import DATASETS_JOB_QUEUE
from renku.service.serializers.datasets import DatasetImportRequest, DatasetImportResponseRPC
from renku.service.views import result_response


class DatasetsImportCtrl(ServiceCtrl, ReadWithSyncOperation):
    """Controller for datasets import endpoint."""

    REQUEST_SERIALIZER = DatasetImportRequest()
    RESPONSE_SERIALIZER = DatasetImportResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a datasets import controller."""
        self.ctx = DatasetsImportCtrl.REQUEST_SERIALIZER.load(request_data)

        if self.ctx.get("commit_message") is None:
            self.ctx["commit_message"] = "service: dataset import of {0}".format(self.ctx["dataset_uri"])

        super(DatasetsImportCtrl, self).__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        job = self.cache.make_job(
            self.user,
            # NOTE: To support operation to be execute on remote project, this behaviour should be updated.
            project=self.ctx["project_id"],
            job_data={"renku_op": "dataset_import", "client_extras": self.ctx.get("client_extras")},
        )

        with enqueue_retry(DATASETS_JOB_QUEUE) as queue:
            queue.enqueue(
                dataset_import,
                self.user_data,
                job.job_id,
                self.ctx["project_id"],
                self.ctx["dataset_uri"],
                name=self.ctx.get("name"),
                extract=self.ctx.get("extract", False),
                job_timeout=int(os.getenv("WORKER_DATASET_JOBS_TIMEOUT", 1800)),
                result_ttl=int(os.getenv("WORKER_DATASET_JOBS_RESULT_TTL", 500)),
            )

        return job

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(DatasetsImportCtrl.RESPONSE_SERIALIZER, self.execute_op())



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service jobs view tests."""
import copy
import uuid
from datetime import datetime

import pytest
from marshmallow.utils import isoformat

from renku.service.cache.models.project import Project


@pytest.mark.service
def test_jobs_view_identity_protected(svc_client):
    """Check serving of user requested jobs."""
    headers = {
        "Content-Type": "application/json",
        "accept": "application/json",
    }
    response = svc_client.get("/jobs", headers=headers)

    assert {"error"} == set(response.json.keys())
    assert "user identification is incorrect or missing" == response.json["error"]["reason"]


@pytest.mark.service
def test_jobs_view_empty_result(svc_client):
    """Check empty result for user requested jobs."""
    headers = {"Content-Type": "application/json", "accept": "application/json", "Renku-User-Id": "user"}
    response = svc_client.get("/jobs", headers=headers)

    assert {"result"} == set(response.json.keys())
    assert [] == response.json["result"]["jobs"]


@pytest.mark.service
def test_jobs_view_expected_job(svc_client_cache):
    """Check non-empty result for user requested job."""
    svc_client, headers, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "user"})
    job_data = {
        "job_id": uuid.uuid4().hex,
        "state": "CREATED",
        "renku_op": "dataset_import",
        "created_at": isoformat(datetime.now()),
        "updated_at": isoformat(datetime.now()),
        "extras": {"progress": 42},
    }

    project = Project(project_id="123", user_id=user.user_id, owner="renkumeister", name="testproject")
    project.abs_path.mkdir(parents=True, exist_ok=True)
    project.save()

    job = cache.make_job(user, project=project, job_data=job_data)

    response = svc_client.get("/jobs", headers=headers)
    assert 1 == len(response.json["result"]["jobs"])
    assert {"job_id", "state", "created_at", "updated_at", "extras", "client_extras", "renku_op", "project"} == set(
        response.json["result"]["jobs"][0].keys()
    )

    cache.invalidate_job(user, job.job_id)
    response = svc_client.get("/jobs", headers=headers)
    assert 0 == len(response.json["result"]["jobs"])


@pytest.mark.service
def test_jobs_view_check_exclusion(svc_client_cache):
    """Check non-empty result for user requested jobs."""
    svc_client, headers, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "user"})
    excluded_user = cache.ensure_user({"user_id": "excluded_user"})

    for _ in range(10):
        job_data = {
            "job_id": uuid.uuid4().hex,
            "state": "CREATED",
            "created_at": isoformat(datetime.utcnow()),
            "updated_at": isoformat(datetime.utcnow()),
            "extras": {"progress": 42},
            "renku_op": "dataset_import",
        }

        project = Project(project_id="123", user_id=user.user_id, owner="renkumeister", name="testproject")
        project.abs_path.mkdir(parents=True, exist_ok=True)
        project.save()

        job1 = cache.make_job(user, project=project, job_data=job_data)
        assert job1

        new_job = copy.deepcopy(job_data)
        new_job["job_id"] = uuid.uuid4().hex

        job2 = cache.make_job(excluded_user, project=project, job_data=new_job)
        assert job2

        assert job1.job_id != job2.job_id

    response = svc_client.get("/jobs", headers=headers)

    assert {"result"} == set(response.json.keys())
    assert 10 == len(response.json["result"]["jobs"])
    for job in response.json["result"]["jobs"]:
        assert {"job_id", "state", "created_at", "updated_at", "extras", "client_extras", "renku_op", "project"} == set(
            job.keys()
        )


@pytest.mark.service
def test_job_details_auth(svc_client):
    """Check authorization for listing a specific job."""
    headers = {
        "Content-Type": "application/json",
        "accept": "application/json",
    }
    response = svc_client.get("/jobs/myjob", headers=headers)

    assert {"error"} == set(response.json.keys())
    assert "user identification is incorrect or missing" == response.json["error"]["reason"]


@pytest.mark.service
def test_job_details_empty(svc_client):
    """Check job details for a user."""
    headers = {"Content-Type": "application/json", "accept": "application/json", "Renku-User-Id": "user"}
    response = svc_client.get("/jobs/myjob", headers=headers)

    assert {"result"} == set(response.json.keys())
    assert response.json["result"] is None


@pytest.mark.service
def test_job_details_2user(svc_client_cache):
    """Check job details for a user."""
    svc_client, headers, cache = svc_client_cache

    user = cache.ensure_user({"user_id": "user"})
    jobs = [
        {
            "job_id": uuid.uuid4().hex,
            "state": "CREATED",
            "created_at": isoformat(datetime.now()),
            "updated_at": isoformat(datetime.now()),
            "extras": {"progress": 42},
            "renku_op": "dataset_import",
        }
        for _ in range(10)
    ]

    project = Project(project_id="123", user_id=user.user_id, owner="renkumeister", name="testproject")
    project.abs_path.mkdir(parents=True, exist_ok=True)
    project.save()

    for job_data in jobs:
        cache.make_job(user, job_data=job_data, project=project)

    user_headers = copy.deepcopy(headers)
    headers["Renku-User-Id"] = "excluded_user"
    excluded_user_headers = copy.deepcopy(headers)

    for job in jobs:
        response = svc_client.get("/jobs/{0}".format(job["job_id"]), headers=user_headers)
        assert response
        assert job["job_id"] == response.json["result"]["job_id"]

        response = svc_client.get("/jobs/{0}".format(job["job_id"]), headers=excluded_user_headers)
        assert response.json["result"] is None



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test execution of commands in parallel."""

import os
import subprocess
import sys
import time
from pathlib import Path


def test_run_in_isolation(runner, project, client, run, subdirectory):
    """Test run in isolation."""
    import filelock

    cwd = Path(project)
    with client.commit():
        with (cwd / ".gitignore").open("a") as f:
            f.write("\nlock")

    prefix = [
        "run",
        "--no-output",
    ]
    cmd = ["python", "-S", "-c", 'import os, sys; sys.exit(1 if os.path.exists("lock") else 0)']

    head = client.repo.head.commit.hexsha

    with filelock.FileLock("lock"):
        assert 1 == run(args=prefix + cmd)
        assert client.repo.head.commit.hexsha == head

        assert 0 == run(prefix + ["--isolation"] + cmd)
        assert client.repo.head.commit.hexsha != head


def test_file_modification_during_run(
    tmpdir, runner, project, client, run, subdirectory, no_lfs_size_limit, no_lfs_warning
):
    """Test run in isolation."""
    script = client.path / "script.py"
    output = client.path / "output"
    lock = Path(str(tmpdir.join("lock")))

    with client.commit():
        script.write_text(
            "import os, time, sys\n"
            'open("{lock}", "a")\n'
            'while os.path.exists("{lock}"):\n'
            "    time.sleep(1)\n"
            "sys.stdout.write(sys.stdin.read())\n"
            "sys.stdout.flush()\n".format(lock=str(lock))
        )

    prefix = [
        sys.executable,
        "-m",
        "renku.cli",
        "run",
        "--isolation",
    ]
    cmd = ["python", os.path.relpath(script, os.getcwd())]

    previous = client.repo.head.commit

    with output.open("wb") as stdout:
        process = subprocess.Popen(prefix + cmd, stdin=subprocess.PIPE, stdout=stdout)

        while not lock.exists() and process.poll() is None:
            time.sleep(1)

        assert process.poll() is None, "Subprocess exited prematurely"

        with script.open("w") as fp:
            fp.write('print("edited")')

        lock.unlink()

        process.communicate(input=b"test")
        assert 0 == process.wait()

    with output.open("r") as fp:
        assert "test" == fp.read().strip()

    diff = previous.diff(client.repo.head.commit)
    modifications = [modification for modification in diff if modification.change_type == "M"]
    assert 0 == len(modifications)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for non-dataset CLI command."""
from pathlib import Path

import pytest
from flaky import flaky

from renku.cli import cli
from renku.core.commands.clone import project_clone
from renku.core.utils.contexts import chdir


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize("url", ["https://dev.renku.ch/gitlab/renku-testing/project-9"])
def test_renku_clone(runner, monkeypatch, url):
    """Test cloning of a Renku repo and existence of required settings."""
    from renku.core.management.storage import StorageApiMixin

    with runner.isolated_filesystem() as project_path:
        result = runner.invoke(cli, ["clone", url, project_path])
        assert 0 == result.exit_code, result.output + str(result.stderr_bytes)
        assert (Path(project_path) / "Dockerfile").exists()

        # Check Git hooks are installed
        result = runner.invoke(cli, ["githooks", "install"])
        assert 0 == result.exit_code, result.output + str(result.stderr_bytes)
        assert "Hook already exists." in result.output

        result = runner.invoke(cli, ["migrate"])
        assert 0 == result.exit_code, result.output + str(result.stderr_bytes)

        # Check Git LFS is enabled
        with monkeypatch.context() as monkey:
            # Pretend that git-lfs is not installed.
            monkey.setattr(StorageApiMixin, "storage_installed", False)
            # Repo is using external storage but it's not installed.
            result = runner.invoke(cli, ["run", "touch", "output"])

            assert "External storage is not configured" in result.output
            assert 1 == result.exit_code, result.output + str(result.stderr_bytes)


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize("url", ["https://dev.renku.ch/gitlab/renku-testing/project-9"])
def test_renku_clone_with_config(tmp_path, url):
    """Test cloning of a Renku repo and existence of required settings."""
    with chdir(tmp_path):
        repo, _ = project_clone(url, config={"user.name": "sam", "user.email": "s@m.i", "filter.lfs.custom": "0"})

        assert "master" == repo.active_branch.name
        reader = repo.config_reader()
        reader.values()

        lfs_config = dict(reader.items("filter.lfs"))
        assert "0" == lfs_config.get("custom")


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize("url", ["https://dev.renku.ch/gitlab/renku-testing/project-9"])
def test_renku_clone_checkout_rev(tmp_path, url):
    """Test cloning of a repo checking out a rev with static config."""
    with chdir(tmp_path):
        repo, _ = project_clone(
            url,
            config={"user.name": "sam", "user.email": "s@m.i", "filter.lfs.custom": "0"},
            checkout_rev="97f907e1a3f992d4acdc97a35df73b8affc917a6",
        )

        assert "97f907e1a3f992d4acdc97a35df73b8affc917a6" == str(repo.active_branch)
        reader = repo.config_reader()
        reader.values()

        lfs_config = dict(reader.items("filter.lfs"))
        assert "0" == lfs_config.get("custom")


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize("rev", ["test-branch", "my-tag",])
def test_renku_clone_checkout_revs(tmp_path, rev):
    """Test cloning of a Renku repo checking out a rev."""
    with chdir(tmp_path):
        repo, _ = project_clone("https://dev.renku.ch/gitlab/contact/no-renku.git", checkout_rev=rev,)

        assert rev == repo.active_branch.name


@pytest.mark.integration
@pytest.mark.parametrize("path,expected_path", [("", "project-9"), (".", "."), ("new-name", "new-name")])
@flaky(max_runs=10, min_passes=1)
def test_renku_clone_uses_project_name(runner, path, expected_path):
    """Test renku clone uses project name as target-path by default."""
    remote = "https://dev.renku.ch/gitlab/renku-testing/project-9"

    with runner.isolated_filesystem() as project_path:
        result = runner.invoke(cli, ["clone", remote, path])
        assert 0 == result.exit_code, result.output + str(result.stderr_bytes)
        assert (Path(project_path) / expected_path / "Dockerfile").exists()


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_renku_clone_private_project_error(runner):
    """Test renku clone prints proper error message when a project is private."""
    remote = "git@dev.renku.ch:mohammad.alisafaee/test-private-project.git"

    with runner.isolated_filesystem():
        result = runner.invoke(cli, ["clone", remote, ""])

        assert 0 != result.exit_code
        assert "Please make sure you have the correct access rights" in result.output
        assert "and the repository exists." in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling a local repository."""
import hashlib
import json
import os
import shutil
import subprocess
import uuid
from collections import defaultdict
from contextlib import contextmanager
from subprocess import check_output

import attr
import filelock
import yaml
from jinja2 import Template
from werkzeug.utils import cached_property, secure_filename

from renku.core import errors
from renku.core.compat import Path
from renku.core.management.config import RENKU_HOME
from renku.core.models.projects import Project
from renku.core.models.refs import LinkReference

from .git import GitCore

DEFAULT_DATA_DIR = "data"


def default_path():
    """Return default repository path."""
    from git import InvalidGitRepositoryError

    from renku.core.commands.git import get_git_home

    try:
        return get_git_home()
    except InvalidGitRepositoryError:
        return "."


def path_converter(path):
    """Converter for path in PathMixin."""
    return Path(path).resolve()


@attr.s
class PathMixin:
    """Define a default path attribute."""

    path = attr.ib(default=default_path, converter=path_converter,)

    @path.validator
    def _check_path(self, _, value):
        """Check the path exists and it is a directory."""
        if not (value.exists() and value.is_dir()):
            raise ValueError("Define an existing directory.")


@attr.s
class RepositoryApiMixin(GitCore):
    """Client for handling a local repository."""

    renku_home = attr.ib(default=RENKU_HOME)
    """Define a name of the Renku folder (default: ``.renku``)."""

    renku_path = attr.ib(init=False)
    """Store a ``Path`` instance of the Renku folder."""

    parent = attr.ib(default=None)
    """Store a pointer to the parent repository."""

    data_dir = attr.ib(
        default=DEFAULT_DATA_DIR, kw_only=True, converter=lambda value: str(value) if value else DEFAULT_DATA_DIR
    )
    """Define a name of the folder for storing datasets."""

    METADATA = "metadata.yml"
    """Default name of Renku config file."""

    LOCK_SUFFIX = ".lock"
    """Default suffix for Renku lock file."""

    WORKFLOW = "workflow"
    """Directory for storing workflow in Renku."""

    ACTIVITY_INDEX = "activity_index.yaml"
    """Caches activities that generated a path."""

    DOCKERFILE = "Dockerfile"
    """Name of the Dockerfile in the repo."""

    TEMPLATE_CHECKSUMS = "template_checksums.json"

    RENKU_PROTECTED_PATHS = [
        "\\.renku/.*",
        "Dockerfile",
        "\\.dockerignore",
        "\\.gitignore",
        "\\.gitattributes",
        "\\.gitlab-ci\\.yml",
        "environment\\.yml",
        "requirements\\.txt",
    ]

    _commit_activity_cache = attr.ib(factory=dict)

    _activity_index = attr.ib(default=None)

    _remote_cache = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        #: Configure Renku path.
        path = Path(self.renku_home)
        if not path.is_absolute():
            path = self.path / path

        path.relative_to(path)
        self.renku_path = path

        data_dir = self.get_value("renku", self.DATA_DIR_CONFIG_KEY, local_only=True)
        self.data_dir = data_dir or self.data_dir

        self._subclients = {}

        self._project = None

        super().__attrs_post_init__()

        # initialize submodules
        if self.repo:
            try:
                check_output(["git", "submodule", "update", "--init", "--recursive"], cwd=str(self.path))
            except subprocess.CalledProcessError:
                pass

    @property
    def latest_agent(self):
        """Returns latest agent version used in the repository."""
        return self.project.agent_version

    @property
    def lock(self):
        """Create a Renku config lock."""
        return filelock.FileLock(str(self.renku_path.with_suffix(self.LOCK_SUFFIX)), timeout=0,)

    @property
    def renku_metadata_path(self):
        """Return a ``Path`` instance of Renku metadata file."""
        return self.renku_path.joinpath(self.METADATA)

    @property
    def workflow_path(self):
        """Return a ``Path`` instance of the workflow folder."""
        return self.renku_path / self.WORKFLOW

    @property
    def activity_index_path(self):
        """Path to the activity filepath cache."""
        return self.renku_path / self.ACTIVITY_INDEX

    @property
    def docker_path(self):
        """Path to the Dockerfile."""
        return self.path / self.DOCKERFILE

    @property
    def template_checksums(self):
        """Return a ``Path`` instance to the template checksums file."""
        return self.renku_path / self.TEMPLATE_CHECKSUMS

    @cached_property
    def cwl_prefix(self):
        """Return a CWL prefix."""
        self.workflow_path.mkdir(parents=True, exist_ok=True)  # for Python 3.5
        return str(self.workflow_path.resolve().relative_to(self.path))

    @property
    def project(self):
        """Return the Project instance."""
        if self.renku_metadata_path.exists() and self._project is None:
            self._project = Project.from_yaml(self.renku_metadata_path, client=self)

        return self._project

    @property
    def remote(self, remote_name="origin"):
        """Return host, owner and name of the remote if it exists."""
        from renku.core.models.git import GitURL

        original_remote_name = remote_name

        if original_remote_name in self._remote_cache:
            return self._remote_cache[original_remote_name]

        host = owner = name = None
        try:
            remote_branch = self.repo.head.reference.tracking_branch()
            if remote_branch is not None:
                remote_name = remote_branch.remote_name
        except TypeError:
            pass

        try:
            url = GitURL.parse(self.repo.remotes[remote_name].url)

            # Remove gitlab. unless running on gitlab.com.
            hostname_parts = url.hostname.split(".")
            if len(hostname_parts) > 2 and hostname_parts[0] == "gitlab":
                hostname_parts = hostname_parts[1:]
            url = attr.evolve(url, hostname=".".join(hostname_parts))
        except IndexError:
            url = None

        if url:
            host = url.hostname
            owner = url.owner
            name = url.name

        remote = {"host": host, "owner": owner, "name": name}
        self._remote_cache[original_remote_name] = remote

        return remote

    def is_project_set(self):
        """Return if project is set for the client."""
        return self._project is not None

    def process_commit(self, commit=None, path=None):
        """Build an :class:`~renku.core.models.provenance.activities.Activity`.

        :param commit: Commit to process. (default: ``HEAD``)
        :param path: Process a specific CWL file.
        """
        from renku.core.models.provenance.activities import Activity

        commit = commit or self.repo.head.commit
        if len(commit.parents) > 1:
            return Activity(commit=commit, client=self)

        if path is None:
            for file_ in commit.stats.files.keys():
                # Find a process (CommandLineTool or Workflow)
                if self.is_workflow(file_):
                    if path is not None:
                        # Regular activity since it edits multiple CWL files
                        return Activity(commit=commit, client=self)

                    path = file_

        if not path:
            # search for activities a file could have been a part of
            activities = self.activities_for_paths(commit.stats.files.keys(), file_commit=commit, revision="HEAD")
            if len(activities) > 1:
                raise errors.CommitProcessingError(
                    "Found multiple activities that produced the same entity at commit {}".format(commit)
                )
            if activities:
                return activities[0]
        else:
            data = (commit.tree / path).data_stream.read()
            process = Activity.from_jsonld(yaml.safe_load(data), client=self, commit=commit)

            return process

        return Activity(commit=commit, client=self)

    def is_workflow(self, path):
        """Check if the path is a valid CWL file."""
        return path.startswith(self.cwl_prefix) and path.endswith(".yaml")

    def find_previous_commit(self, paths, revision="HEAD", return_first=False, full=False):
        """Return a previous commit for a given path starting from ``revision``.

        :param revision: revision to start from, defaults to ``HEAD``
        :param return_first: show the first commit in the history
        :param full: return full history
        :raises KeyError: if path is not present in the given commit
        """
        kwargs = {}

        if full:
            kwargs["full_history"] = True

        if return_first:
            file_commits = list(self.repo.iter_commits(revision, paths=paths, **kwargs))
        else:
            file_commits = list(self.repo.iter_commits(revision, paths=paths, max_count=1, **kwargs))

        if not file_commits:
            raise KeyError("Could not find a file {0} in range {1}".format(paths, revision))

        return file_commits[-1 if return_first else 0]

    @cached_property
    def workflow_names(self):
        """Return index of workflow names."""
        names = defaultdict(list)
        for ref in LinkReference.iter_items(self, common_path="workflows"):
            names[str(ref.reference.relative_to(self.path))].append(ref.name)
        return names

    @cached_property
    def submodules(self):
        """Return list of submodules it belongs to."""
        if self.parent:
            client, submodule = self.parent
            return client.submodules + [submodule.name]
        return []

    def subclients(self, parent_commit):
        """Return mapping from submodule to client."""
        if parent_commit in self._subclients:
            return self._subclients[parent_commit]

        try:
            from git import Submodule

            submodules = [submodule for submodule in Submodule.iter_items(self.repo, parent_commit=parent_commit)]
        except (RuntimeError, ValueError):
            # There are no submodules associated with the given commit.
            submodules = []

        subclients = {}
        for submodule in submodules:
            subpath = (self.path / submodule.path).resolve()
            is_renku = subpath / Path(self.renku_home)

            if subpath.exists() and is_renku.exists():
                subclients[submodule] = self.__class__(path=subpath, parent=(self, submodule),)

        return subclients

    def resolve_in_submodules(self, commit, path):
        """Resolve filename in submodules."""
        original_path = self.path / path
        in_vendor = str(path).startswith(".renku/vendors")

        if original_path.is_symlink() or in_vendor:
            original_path = Path(os.path.realpath(os.path.abspath(str(original_path))))

            for submodule, subclient in self.subclients(commit).items():
                if (Path(submodule.path) / Path(".git")).exists():

                    try:
                        subpath = original_path.relative_to(subclient.path)
                        return (
                            subclient,
                            subclient.find_previous_commit(subpath, revision=submodule.hexsha),
                            subpath,
                        )
                    except ValueError:
                        pass

        return self, commit, path

    @contextmanager
    def with_commit(self, commit):
        """Yield the state of the repo at a specific commit."""
        current_branch = None
        current_commit = None

        try:
            current_branch = self.repo.active_branch
        except TypeError as e:
            # not on a branch, detached head
            if "HEAD is a detached" in str(e):
                current_commit = self.repo.head.commit
            else:
                raise ValueError("Couldn't get active branch or commit", e)

        self.repo.git.checkout(commit)

        try:
            yield
        finally:
            if current_branch:
                self.repo.git.checkout(current_branch)
            elif current_commit:
                self.repo.git.checkout(current_commit)

    @contextmanager
    def with_metadata(self, read_only=False, name=None):
        """Yield an editable metadata object."""
        metadata_path = self.renku_metadata_path

        if metadata_path.exists():
            metadata = Project.from_yaml(metadata_path, client=self)
        else:
            metadata = Project(name=name, client=self)

        yield metadata

        if not read_only:
            metadata.to_yaml(path=metadata_path)

    @contextmanager
    def with_workflow_storage(self):
        """Yield a workflow storage."""
        from renku.core.models.cwl.workflow import Workflow

        workflow = Workflow()
        yield workflow

        for step in workflow.steps:
            step_name = "{0}_{1}.yaml".format(uuid.uuid4().hex, secure_filename("_".join(step.run.baseCommand)),)

            workflow_path = self.workflow_path
            if not workflow_path.exists():
                workflow_path.mkdir()

            path = workflow_path / step_name

            run = step.run.generate_process_run(client=self, commit=self.repo.head.commit, path=path,)
            run.to_yaml(path=path)
            self.add_to_activity_index(run)

    def init_repository(self, force=False, user=None):
        """Initialize an empty Renku repository."""
        from git import Repo

        from renku.core.models.provenance.agents import Person

        # verify if folder is empty
        if self.repo is not None and not force:
            raise errors.InvalidFileOperation(
                "Folder {0} already contains file. Use --force to overwrite".format(self.repo.git_dir)
            )

        # initialize repo and set user data
        path = self.path.absolute()
        self.repo = Repo.init(str(path))
        if user:
            config_writer = self.repo.config_writer()
            for key, value in user.items():
                config_writer.set_value("user", key, value)
            config_writer.release()

        # verify if author information is available
        Person.from_git(self.repo)

    @property
    def path_activity_cache(self):
        """Cache of all activities and their generated paths."""
        # TODO: this is there for performance reasons. Remove once graph
        # is stored as a flat, append-only list (Should be graph query
        # in the future)
        if self._activity_index:
            return self._activity_index

        path = self.activity_index_path

        if path.exists():
            with path.open("r") as stream:
                self._activity_index = yaml.safe_load(stream)
        else:
            self._activity_index = {}

        return self._activity_index

    def add_to_activity_index(self, activity):
        """Add an activity and it's generations to the cache."""
        for g in activity.generated:
            if g.path not in self.path_activity_cache:
                self.path_activity_cache[g.path] = {}
            hexsha = g.commit.hexsha
            if hexsha not in self.path_activity_cache[g.path]:
                self.path_activity_cache[g.path][hexsha] = []

            if activity.path in self.path_activity_cache[g.path][hexsha]:
                continue

            self.path_activity_cache[g.path][g.commit.hexsha].append(activity.path)

        if self.path_activity_cache:
            with self.activity_index_path.open("w") as stream:
                yaml.dump(self.path_activity_cache, stream)

    def activities_for_paths(self, paths, file_commit=None, revision="HEAD"):
        """Get all activities involving a path."""
        from renku.core.models.provenance.activities import Activity

        result = set()

        for p in paths:
            if p not in self.path_activity_cache:
                parent_paths = [a for a in self.path_activity_cache.keys() if p.startswith(a)]
                if not parent_paths:
                    continue
                matching_activities = [
                    a
                    for k, v in self.path_activity_cache.items()
                    for ck, cv in v.items()
                    for a in cv
                    if k in parent_paths and (not file_commit or ck == file_commit.hexsha)
                ]
            else:
                matching = self.path_activity_cache[p]
                if file_commit:
                    if file_commit.hexsha not in matching:
                        continue
                    matching_activities = matching[file_commit.hexsha]
                else:
                    matching_activities = [a for v in matching.values() for a in v]

            for a in matching_activities:
                if (a, revision) in self._commit_activity_cache:
                    activity = self._commit_activity_cache[(a, revision)]
                else:
                    try:
                        commit = self.find_previous_commit(a, revision=revision)
                    except KeyError:
                        continue

                    activity = Activity.from_yaml(self.path / a, client=self, commit=commit)

                    self._commit_activity_cache[(a, revision)] = activity
                result.add(activity)

        return list(result)

    def import_from_template(self, template_path, metadata, force=False):
        """Render template files from a template directory."""
        checksums = {}
        for file in template_path.glob("**/*"):
            rel_path = file.relative_to(template_path)
            destination = self.path / rel_path
            try:
                # TODO: notify about the expected variables - code stub:
                # with file.open() as fr:
                #     file_content = fr.read()
                #     # look for the required keys
                #     env = Environment()
                #     parsed = env.parse(file_content)
                #     variables = meta.find_undeclared_variables(parsed)

                # parse file and process it
                template = Template(file.read_text())
                rendered_content = template.render(metadata)
                # NOTE: the path could contain template variables, we need to template it
                destination = Path(Template(str(destination)).render(metadata))
                destination.write_text(rendered_content)
                checksums[str(rel_path)] = self._content_hash(destination)
            except IsADirectoryError:
                destination.mkdir(parents=True, exist_ok=True)
            except TypeError:
                shutil.copy(file, destination)

        self.template_checksums.parent.mkdir(parents=True, exist_ok=True)

        with open(self.template_checksums, "w") as checksum_file:
            json.dump(checksums, checksum_file)

    def check_immutable_template_files(self, *paths):
        """Check paths and return a list of those that are marked immutable in the project template."""
        if not self.project.immutable_template_files:
            return []

        immutable_template_files = self.project.immutable_template_files or []
        return [p for p in paths if str(p) in immutable_template_files]

    def _content_hash(self, path):
        """Calculate the sha256 hash of a file."""
        sha256_hash = hashlib.sha256()
        with open(str(path), "rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service migrate project controller."""
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadWithSyncOperation
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.project import execute_migration, migrate_job
from renku.service.jobs.queues import MIGRATIONS_JOB_QUEUE
from renku.service.serializers.cache import (
    ProjectMigrateAsyncResponseRPC,
    ProjectMigrateRequest,
    ProjectMigrateResponseRPC,
)
from renku.service.views import result_response


class MigrateProjectCtrl(ServiceCtrl, ReadWithSyncOperation):
    """Controller for migrating project endpoint."""

    REQUEST_SERIALIZER = ProjectMigrateRequest()
    RESPONSE_SERIALIZER = ProjectMigrateResponseRPC()
    JOB_RESPONSE_SERIALIZER = ProjectMigrateAsyncResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct controller."""
        self.ctx = MigrateProjectCtrl.REQUEST_SERIALIZER.load(request_data)

        self.force_template_update = self.ctx.get("force_template_update", False)
        self.skip_template_update = self.ctx.get("skip_template_update", False)
        self.skip_docker_update = self.ctx.get("skip_docker_update", False)
        self.skip_migrations = self.ctx.get("skip_migrations", False)
        self.commit_message = self.ctx.get("commit_message", None)
        super(MigrateProjectCtrl, self).__init__(cache, user_data, request_data)

        self.project = self.cache.get_project(self.user, self.ctx["project_id"])

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        messages, was_migrated, template_migrated, docker_migrated = execute_migration(
            self.project,
            self.force_template_update,
            self.skip_template_update,
            self.skip_docker_update,
            self.skip_migrations,
            self.commit_message,
        )
        response = {
            "messages": messages,
            "was_migrated": was_migrated,
            "template_migrated": template_migrated,
            "docker_migrated": docker_migrated,
        }

        if was_migrated or template_migrated or docker_migrated:
            response["remote_branch"] = self.sync()

        return response

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        if self.ctx.get("is_delayed", False):
            job = self.cache.make_job(
                self.user,
                project=self.project,
                job_data={"renku_op": "migrate_job", "client_extras": self.ctx.get("client_extras")},
            )

            with enqueue_retry(MIGRATIONS_JOB_QUEUE) as queue:
                queue.enqueue(
                    migrate_job, self.user_data, self.project.project_id, job.job_id, self.commit_message,
                )

            return result_response(MigrateProjectCtrl.JOB_RESPONSE_SERIALIZER, job)

        return result_response(MigrateProjectCtrl.RESPONSE_SERIALIZER, self.execute_op())



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project related jobs."""

from git import GitCommandError, Repo
from urllib3.exceptions import HTTPError

from renku.core.commands.migrate import migrate_project
from renku.core.commands.save import repo_sync
from renku.core.errors import ParameterError, RenkuException
from renku.core.utils.contexts import chdir
from renku.service.logger import worker_log
from renku.service.views.decorators import requires_cache


def execute_migration(
    project, force_template_update, skip_template_update, skip_docker_update, skip_migrations, commit_message
):
    """Execute project migrations."""
    messages = []
    worker_log.debug(f"migrating {project.abs_path}")

    def collect_message(msg):
        """Collect migration message."""
        messages.append(msg)

    with chdir(project.abs_path):
        was_migrated, template_migrated, docker_migrated = migrate_project(
            progress_callback=collect_message,
            force_template_update=force_template_update,
            skip_template_update=skip_template_update,
            skip_docker_update=skip_docker_update,
            skip_migrations=skip_migrations,
            commit_message=commit_message,
        )

    worker_log.debug(f"migration finished - was_migrated={was_migrated}")
    return messages, was_migrated, template_migrated, docker_migrated


@requires_cache
def migrate_job(
    cache,
    user_data,
    project_id,
    user_job_id,
    force_template_update,
    skip_template_update,
    skip_docker_update,
    skip_migrations,
    commit_message,
):
    """Execute migrations job."""
    user = cache.ensure_user(user_data)
    worker_log.debug(f"executing dataset import job for {user.user_id}:{user.fullname}")

    user_job = cache.get_job(user, user_job_id)

    try:
        project = cache.get_project(user, project_id)
        messages, was_migrated, template_migrated, docker_migrated = execute_migration(
            project, force_template_update, skip_template_update, skip_docker_update, skip_migrations, commit_message
        )

        user_job.update_extras("messages", messages)
        user_job.update_extras("was_migrated", was_migrated)
        user_job.update_extras("template_migrated", template_migrated)
        user_job.update_extras("docker_migrated", docker_migrated)

        worker_log.debug("operation successful - syncing with remote")
        _, remote_branch = repo_sync(Repo(project.abs_path), remote="origin")
        user_job.update_extras("remote_branch", remote_branch)

        user_job.complete()
        worker_log.debug("job completed")
    except (HTTPError, ParameterError, GitCommandError, RenkuException) as exp:
        user_job.update_extras("error", str(exp))
        user_job.fail_job()

        # Reraise exception, so we see trace in job metadata
        # and in metrics as failed job.
        raise exp



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for integration with Calamus."""

import copy
import inspect
from datetime import datetime, timezone

import marshmallow
from calamus import fields
from calamus.schema import JsonLDSchema as CalamusJsonLDSchema
from calamus.utils import normalize_type, normalize_value
from marshmallow.base import SchemaABC

prov = fields.Namespace("http://www.w3.org/ns/prov#")
rdfs = fields.Namespace("http://www.w3.org/2000/01/rdf-schema#")
renku = fields.Namespace("https://swissdatasciencecenter.github.io/renku-ontology#")
schema = fields.Namespace("http://schema.org/")
wfprov = fields.Namespace("http://purl.org/wf4ever/wfprov#")
oa = fields.Namespace("http://www.w3.org/ns/oa#")
dcterms = fields.Namespace("http://purl.org/dc/terms/")


class JsonLDSchema(CalamusJsonLDSchema):
    """Base schema class for Renku."""

    def __init__(self, *args, commit=None, client=None, **kwargs):
        """Create an instance."""
        self._commit = commit
        self._client = client
        super().__init__(*args, **kwargs)

    def _deserialize(self, *args, **kwargs):
        data = super()._deserialize(*args, **kwargs)
        const_args = inspect.signature(self.opts.model)
        parameters = const_args.parameters.values()

        if any(p.name == "client" for p in parameters):
            self._add_field_to_data(data, "client", self._client)

        if any(p.name == "commit" for p in parameters):
            if self._commit:
                self._add_field_to_data(data, "commit", self._commit)
            elif (
                self._client
                and "_label" in data
                and data["_label"]
                and "@UNCOMMITTED" not in data["_label"]
                and "@" in data["_label"]
            ):
                try:
                    self._add_field_to_data(data, "commit", self._client.repo.commit(data["_label"].rsplit("@")[1]))
                except ValueError as e:
                    if "could not be resolved, git returned" not in str(e):
                        raise
                    # NOTE: This means the commit does not exist in the local repository. Could be an external file?

        return data

    def _add_field_to_data(self, data, name, value):
        if value:
            if name in data:
                raise ValueError(f"Field {name} is already in data {data}")
            data[name] = value

    def _fix_timezone(self, value):
        """Fix timezone of non-aware datetime objects."""
        if isinstance(value, datetime) and not value.tzinfo:
            # set timezone to local timezone
            tz = datetime.now(timezone.utc).astimezone().tzinfo
            value = value.replace(tzinfo=tz)
        return value


class Uri(fields._JsonLDField, marshmallow.fields.String, marshmallow.fields.Dict):
    """A Dict/String field."""

    def __init__(self, *args, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, str):
            value = super(fields._JsonLDField, self)._serialize(value, attr, obj, **kwargs)
            if self.parent.opts.add_value_types:
                value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#string"}
        elif isinstance(value, dict):
            value = super(marshmallow.fields.String, self)._serialize(value, attr, obj, **kwargs)

        return value

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)
        if not value:
            return None
        elif isinstance(value, str):
            return value
        elif isinstance(value, dict):
            return super(marshmallow.fields.String, self)._deserialize(value, attr, data, **kwargs)
        else:
            raise ValueError("Invalid type for field {}: {}".format(self.name, type(value)))


class StringList(fields.String):
    """A String field that might be a list when deserializing."""

    def __init__(self, *args, return_max_value=True, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)
        self.return_max_value = return_max_value

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)

        if isinstance(value, (list, tuple, set)):
            value = sorted(value, reverse=self.return_max_value)
            value = value[0] if len(value) > 0 else None

        return super()._deserialize(value, attr, data, **kwargs)


class DateTimeList(fields.DateTime):
    """A DateTime field that might be a list when deserializing."""

    def __init__(self, *args, **kwargs):
        """Create an instance."""
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        value = normalize_value(value)

        if isinstance(value, (list, tuple, set)):
            value = sorted(value)
            value = value[0] if len(value) > 0 else None

        return super()._deserialize(value, attr, data, **kwargs)


class Nested(fields.Nested):
    """Nested field that passes along client and commit info."""

    def __init__(self, *args, propagate_client=True, **kwargs):
        """Init method."""
        super().__init__(*args, **kwargs)

        self.propagate_client = propagate_client

    @property
    def schema(self):
        """The nested ``calamus.Schema`` object.

        This method was copied from marshmallow and modified to support
        multiple different nested schemes.
        """
        if not self._schema:
            # Inherit context from parent.
            context = getattr(self.parent, "context", {})
            self._schema = {"from": {}, "to": {}}
            for nest in self.nested:
                if isinstance(nest, SchemaABC):
                    rdf_type = str(normalize_type(nest.opts.rdf_type))
                    model = nest.opts.model
                    if not rdf_type or not model:
                        raise ValueError("Both rdf_type and model need to be set on the " "schema for nested to work")
                    _schema = copy.copy(nest)
                    _schema.context.update(context)
                    # Respect only and exclude passed from parent and
                    # re-initialize fields
                    set_class = _schema.set_class
                    if self.only is not None:
                        if self._schema.only is not None:
                            original = _schema.only
                        else:  # only=None -> all fields
                            original = _schema.fields.keys()
                        _schema.only = set_class(self.only).intersection(original)
                    if self.exclude:
                        original = _schema.exclude
                        _schema.exclude = set_class(self.exclude).union(original)
                    _schema._init_fields()
                    _schema._visited = self.root._visited
                    self._schema["from"][rdf_type] = _schema
                    self._schema["to"][model] = _schema
                else:
                    if isinstance(nest, type) and issubclass(nest, SchemaABC):
                        schema_class = nest
                    elif not isinstance(nest, (str, bytes)):
                        raise ValueError("Nested fields must be passed a " "Schema, not {}.".format(nest.__class__))
                    elif nest == "self":
                        ret = self
                        while not isinstance(ret, SchemaABC):
                            ret = ret.parent
                        schema_class = ret.__class__
                    else:
                        schema_class = marshmallow.class_registry.get_class(nest)

                    rdf_type = str(normalize_type(schema_class.opts.rdf_type))
                    model = schema_class.opts.model
                    if not rdf_type or not model:
                        raise ValueError("Both rdf_type and model need to be set on the " "schema for nested to work")

                    kwargs = {}

                    if self.propagate_client and hasattr(self.root, "_client") and JsonLDSchema in schema_class.mro():
                        kwargs = {"client": self.root._client}

                    self._schema["from"][rdf_type] = schema_class(
                        many=False,
                        only=self.only,
                        exclude=self.exclude,
                        context=context,
                        load_only=self._nested_normalized_option("load_only"),
                        dump_only=self._nested_normalized_option("dump_only"),
                        lazy=self.root.lazy,
                        flattened=self.root.flattened,
                        _visited=self.root._visited,
                        **kwargs,
                    )
                    self._schema["to"][model] = self._schema["from"][rdf_type]
        return self._schema



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets add controller."""
import json
import os
from pathlib import Path

from renku.core.commands.dataset import add_file
from renku.core.errors import RenkuException
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadWithSyncOperation
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.datasets import dataset_add_remote_file
from renku.service.jobs.queues import DATASETS_JOB_QUEUE
from renku.service.serializers.datasets import DatasetAddRequest, DatasetAddResponseRPC
from renku.service.views import result_response


class DatasetsAddFileCtrl(ServiceCtrl, ReadWithSyncOperation):
    """Controller for datasets add endpoint."""

    REQUEST_SERIALIZER = DatasetAddRequest()
    RESPONSE_SERIALIZER = DatasetAddResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a datasets add controller."""
        self.ctx = DatasetsAddFileCtrl.REQUEST_SERIALIZER.load(request_data)

        if not self.ctx["commit_message"]:
            self.ctx["commit_message"] = "service: dataset add {0}".format(self.ctx["name"])

        super(DatasetsAddFileCtrl, self).__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def prepare_paths(self):
        """Prepare local paths which can be added in same transaction."""
        local_paths, enqueued_paths = [], []

        for _file in self.ctx["files"]:
            local_path = None

            if "file_url" in _file:
                commit_message = "{0}{1}".format(self.ctx["commit_message"], _file["file_url"])

                job = self.cache.make_job(
                    self.user,
                    # NOTE: To support operation to be executed on remote project, this behaviour should be updated.
                    project=self.ctx["project_id"],
                    job_data={"renku_op": "dataset_add_remote_file", "client_extras": self.ctx.get("client_extras")},
                )
                _file["job_id"] = job.job_id

                with enqueue_retry(DATASETS_JOB_QUEUE) as queue:
                    queue.enqueue(
                        dataset_add_remote_file,
                        self.user_data,
                        job.job_id,
                        # NOTE: To support operation to be executed on remote project, this behaviour should be updated.
                        self.ctx["project_id"],
                        self.ctx["create_dataset"],
                        commit_message,
                        self.ctx["name"],
                        _file["file_url"],
                        job_timeout=int(os.getenv("WORKER_DATASET_JOBS_TIMEOUT", 1800)),
                        result_ttl=int(os.getenv("WORKER_DATASET_JOBS_RESULT_TTL", 500)),
                    )
                    enqueued_paths.append(_file["file_url"])

                continue

            if "file_id" in _file:
                file = self.cache.get_file(self.user, _file["file_id"])
                local_path = file.abs_path

            elif "file_path" in _file:
                local_path = self.project_path / Path(_file["file_path"])

            if not local_path or not local_path.exists():
                raise RenkuException("invalid file reference: {0}".format(json.dumps(_file)))

            self.ctx["commit_message"] += " {0}".format(local_path.name)
            local_paths.append(str(local_path))

        return local_paths, enqueued_paths

    def renku_op(self):
        """Renku operation for the controller."""
        local_paths, enqueued_paths = self.prepare_paths()

        if not local_paths and not enqueued_paths:
            raise RenkuException("no files processed")

        if local_paths:
            add_file(
                local_paths,
                self.ctx["name"],
                create=self.ctx["create_dataset"],
                force=self.ctx["force"],
                commit_message=self.ctx["commit_message"],
            )

        return local_paths, enqueued_paths

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        op_result, remote_branch = self.execute_and_sync()
        local_paths, enqueued_paths = op_result

        response = self.ctx
        response["remote_branch"] = remote_branch

        return result_response(DatasetsAddFileCtrl.RESPONSE_SERIALIZER, response)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Communicator classes for printing output."""

from functools import wraps
from threading import RLock

from werkzeug.local import Local


class CommunicationCallback:
    """Base communication callback class."""

    lock = RLock()

    def echo(self, msg):
        """Write a message."""

    def info(self, msg):
        """Write an info message."""

    def warn(self, msg):
        """Write a warning message."""

    def error(self, msg):
        """Write an error message."""

    def confirm(self, msg, abort=False):
        """Get confirmation for an action."""

    def start_progress(self, name, total, **kwargs):
        """Create a new progress tracker."""

    def update_progress(self, name, amount):
        """Update a progress tracker."""

    def finalize_progress(self, name):
        """End a progress tracker."""


class _CommunicationManger(CommunicationCallback):
    """Manages all communication callback objects."""

    def __init__(self):
        super().__init__()
        self._listeners = []

    def subscribe(self, listener):
        """Add a new listener for communications."""
        with CommunicationCallback.lock:
            if listener not in self._listeners:
                self._listeners.append(listener)

    def unsubscribe(self, listener):
        """Remove a communication listener."""
        with CommunicationCallback.lock:
            if listener in self._listeners:
                self._listeners.remove(listener)

    def echo(self, msg):
        """Write a message."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.echo(msg)

    def info(self, msg):
        """Write an info message."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.info(msg)

    def warn(self, msg):
        """Write a warning message."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.warn(msg)

    def error(self, msg):
        """Write an error message."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.error(msg)

    def confirm(self, msg, abort=False):
        """Get confirmation for an action."""
        with CommunicationCallback.lock:
            result = True
            for listener in self._listeners:
                result = result and listener.confirm(msg, abort)
            return result

    def start_progress(self, name, total, **kwargs):
        """Create a new progress tracker."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.start_progress(name, total, **kwargs)

    def update_progress(self, name, amount):
        """Update a progress tracker."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.update_progress(name, amount)

    def finalize_progress(self, name):
        """End a progress tracker."""
        with CommunicationCallback.lock:
            for listener in self._listeners:
                listener.finalize_progress(name)


_thread_local = Local()


def ensure_manager(f):
    """Decorator to add communication manager to local thread storage."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        if getattr(_thread_local, "communication_manager", None) is None:
            _thread_local.communication_manager = _CommunicationManger()

        return f(*args, **kwargs)

    return wrapper


@ensure_manager
def subscribe(listener):
    """Subscribe a communication listener."""
    _thread_local.communication_manager.subscribe(listener)


@ensure_manager
def unsubscribe(listener):
    """Unsubscribe a communication listener."""
    _thread_local.communication_manager.unsubscribe(listener)


@ensure_manager
def echo(msg):
    """Write a message to all listeners."""
    _thread_local.communication_manager.echo(msg)


@ensure_manager
def info(msg):
    """Write an info message to all listeners."""
    _thread_local.communication_manager.info(msg)


@ensure_manager
def warn(msg):
    """Write a warning message to all listeners."""
    _thread_local.communication_manager.warn(msg)


@ensure_manager
def error(msg):
    """Write an info message to all listeners."""
    _thread_local.communication_manager.error(msg)


@ensure_manager
def confirm(msg, abort=False):
    """Get confirmation for an action from all listeners."""
    _thread_local.communication_manager.confirm(msg, abort)


@ensure_manager
def start_progress(name, total, **kwargs):
    """Start a progress tracker on all listeners."""
    _thread_local.communication_manager.start_progress(name, total, **kwargs)


@ensure_manager
def update_progress(name, amount):
    """Update a progress tracker on all listeners."""
    _thread_local.communication_manager.update_progress(name, amount)


@ensure_manager
def finalize_progress(name):
    """End a progress tracker on all listeners."""
    _thread_local.communication_manager.finalize_progress(name)


__all__ = [
    "CommunicationCallback",
    "subscribe",
    "unsubscribe",
    "info",
    "warn",
    "error",
    "start_progress",
    "update_progress",
    "finalize_progress",
]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test command builder."""

import threading

import pytest

from renku.cli.utils.callback import ClickCallback
from renku.core.incubation.command import Command
from renku.core.utils import communication
from renku.service.utils.callback import ServiceCallback


def test_communicator_is_unsubscribed(tmp_path):
    """Test communicator is unsubscribed when command is executed."""
    communicator = ServiceCallback()

    command = Command().command(lambda _: communication.echo("Hello world!"))
    command.with_communicator(communicator).build().execute()

    assert ["Hello world!"] == communicator.messages

    communication.echo("More messages.")

    assert ["Hello world!"] == communicator.messages


def test_multi_communicators(tmp_path):
    """Test subscribing multiple communicators."""
    communicator_1 = ServiceCallback()
    communicator_2 = ServiceCallback()

    command = Command().command(lambda _: communication.echo("Hello world!"))
    command.with_communicator(communicator_1).with_communicator(communicator_2).build().execute()

    assert ["Hello world!"] == communicator_1.messages
    assert ["Hello world!"] == communicator_2.messages


def test_multi_threaded_communication(tmp_path):
    """Test communication with multi-threading."""

    def thread_function(name, communicator):
        command = Command().command(lambda _: communication.echo(f"Hello world from {name}!"))
        command.with_communicator(communicator).build().execute()

    communicator_1 = ServiceCallback()
    thread_1 = threading.Thread(target=thread_function, args=("thread-1", communicator_1))

    communicator_2 = ServiceCallback()
    thread_2 = threading.Thread(target=thread_function, args=("thread-2", communicator_2))

    thread_1.start()
    thread_2.start()
    thread_1.join()
    thread_2.join()

    assert ["Hello world from thread-1!"] == communicator_1.messages
    assert ["Hello world from thread-2!"] == communicator_2.messages


def test_click_callback(capsys):
    """Test communication with ClickCallback."""
    communicator = ClickCallback()
    command = Command().command(lambda _: communication.echo("Hello world!"))
    command.with_communicator(communicator).build().execute()

    assert "Hello world!" in capsys.readouterr().out


@pytest.mark.serial
def test_no_callback(capsys):
    """Test communication does not print anything without a callback."""
    command = Command().command(lambda _: communication.echo("Hello world!"))
    command.build().execute()

    assert "Hello world!" not in capsys.readouterr().out
